"""
Tests the localfs tokens interface.
"""

import os

import pytest

import salt.exceptions
import salt.tokens.localfs
import salt.utils.files
from tests.support.mock import patch


class CalledWith:
    def __init__(self, func, called_with=None):
        self.func = func
        if called_with is None:
            self.called_with = []
        else:
            self.called_with = called_with

    def __call__(self, *args, **kwargs):
        self.called_with.append((args, kwargs))
        return self.func(*args, **kwargs)


@pytest.fixture
def expected_data():
    # Default expected data
    return {"this": "is", "some": "token data"}


def test_write_token(tmp_path):
    """
    Validate tokens put in place with an atomic move
    """
    opts = {"token_dir": str(tmp_path)}
    fopen = CalledWith(salt.utils.files.fopen)
    rename = CalledWith(os.rename)
    with patch("salt.utils.files.fopen", fopen), patch("os.rename", rename):
        tdata = salt.tokens.localfs.mk_token(opts, {})
    assert "token" in tdata
    t_path = os.path.join(str(tmp_path), tdata["token"])
    temp_t_path = "{}.tmp".format(t_path)
    assert len(fopen.called_with) == 1, len(fopen.called_with)
    assert fopen.called_with == [((temp_t_path, "w+b"), {})], fopen.called_with
    assert len(rename.called_with) == 1, len(rename.called_with)
    assert rename.called_with == [((temp_t_path, t_path), {})], rename.called_with


def test_get_token_should_return_token_if_exists(tmp_path, expected_data):
    opts = {"token_dir": str(tmp_path)}
    tok = salt.tokens.localfs.mk_token(
        opts=opts,
        tdata=expected_data,
    )["token"]
    actual_data = salt.tokens.localfs.get_token(opts=opts, tok=tok)
    assert expected_data == actual_data


def test_get_token_should_raise_SaltDeserializationError_if_token_file_is_empty(
    tmp_path, expected_data
):
    opts = {"token_dir": str(tmp_path)}
    tok = salt.tokens.localfs.mk_token(
        opts=opts,
        tdata=expected_data,
    )["token"]
    with salt.utils.files.fopen(os.path.join(str(tmp_path), tok), "w") as f:
        f.truncate()
    with pytest.raises(salt.exceptions.SaltDeserializationError) as e:
        salt.tokens.localfs.get_token(opts=opts, tok=tok)


def test_get_token_should_raise_SaltDeserializationError_if_token_file_is_malformed(
    tmp_path, expected_data
):
    opts = {"token_dir": str(tmp_path)}
    tok = salt.tokens.localfs.mk_token(
        opts=opts,
        tdata=expected_data,
    )["token"]
    with salt.utils.files.fopen(os.path.join(str(tmp_path), tok), "w") as f:
        f.truncate()
        f.write("this is not valid msgpack data")
    with pytest.raises(salt.exceptions.SaltDeserializationError) as e:
        salt.tokens.localfs.get_token(opts=opts, tok=tok)



import copy
import logging
import os

import pytest

import salt.ext.tornado
import salt.ext.tornado.gen
import salt.ext.tornado.testing
import salt.minion
import salt.syspaths
import salt.utils.crypt
import salt.utils.event as event
import salt.utils.jid
import salt.utils.platform
import salt.utils.process
from salt._compat import ipaddress
from salt.exceptions import SaltClientError, SaltMasterUnresolvableError, SaltSystemExit
from tests.support.mock import MagicMock, patch

log = logging.getLogger(__name__)


def test_minion_load_grains_false(minion_opts):
    """
    Minion does not generate grains when load_grains is False
    """
    minion_opts["grains"] = {"foo": "bar"}
    with patch("salt.loader.grains") as grainsfunc:
        minion = salt.minion.Minion(minion_opts, load_grains=False)
        assert minion.opts["grains"] == minion_opts["grains"]
        grainsfunc.assert_not_called()


def test_minion_load_grains_true(minion_opts):
    """
    Minion generates grains when load_grains is True
    """
    with patch("salt.loader.grains") as grainsfunc:
        minion = salt.minion.Minion(minion_opts, load_grains=True)
        assert minion.opts["grains"] != {}
        grainsfunc.assert_called()


def test_minion_load_grains_default(minion_opts):
    """
    Minion load_grains defaults to True
    """
    with patch("salt.loader.grains") as grainsfunc:
        minion = salt.minion.Minion(minion_opts)
        assert minion.opts["grains"] != {}
        grainsfunc.assert_called()


@pytest.mark.parametrize(
    "event",
    [
        (
            "fire_event",
            lambda data, tag, cb=None, timeout=60: True,
        ),
        (
            "fire_event_async",
            lambda data, tag, cb=None, timeout=60: salt.ext.tornado.gen.maybe_future(
                True
            ),
        ),
    ],
)
def test_send_req_fires_completion_event(event, minion_opts):
    event_enter = MagicMock()
    event_enter.send.side_effect = event[1]
    event = MagicMock()
    event.__enter__.return_value = event_enter

    with patch("salt.utils.event.get_event", return_value=event):
        minion_opts["random_startup_delay"] = 0
        minion_opts["return_retry_tries"] = 30
        minion_opts["grains"] = {}
        with patch("salt.loader.grains"):
            minion = salt.minion.Minion(minion_opts)

            load = {"load": "value"}
            timeout = 60

            # XXX This is buggy because "async" in event[0] will never evaluate
            # to True and if it *did* evaluate to true the test would fail
            # because you Mock isn't a co-routine.
            if "async" in event[0]:
                rtn = minion._send_req_async(load, timeout).result()
            else:
                rtn = minion._send_req_sync(load, timeout)

            # get the
            for idx, call in enumerate(event.mock_calls, 1):
                if "fire_event" in call[0]:
                    condition_event_tag = (
                        len(call.args) > 1
                        and call.args[1]
                        == f"__master_req_channel_payload/{minion_opts['master']}"
                    )
                    condition_event_tag_error = "{} != {}; Call(number={}): {}".format(
                        idx, call, call.args[1], "__master_req_channel_payload"
                    )
                    condition_timeout = (
                        len(call.kwargs) == 1 and call.kwargs["timeout"] == timeout
                    )
                    condition_timeout_error = "{} != {}; Call(number={}): {}".format(
                        idx, call, call.kwargs["timeout"], timeout
                    )

                    fire_event_called = True
                    assert condition_event_tag, condition_event_tag_error
                    assert condition_timeout, condition_timeout_error

            assert fire_event_called
            assert rtn


async def test_send_req_async_regression_62453(minion_opts):
    event_enter = MagicMock()
    event_enter.send.side_effect = (
        lambda data, tag, cb=None, timeout=60: salt.ext.tornado.gen.maybe_future(True)
    )
    event = MagicMock()
    event.__enter__.return_value = event_enter

    minion_opts["random_startup_delay"] = 0
    minion_opts["return_retry_tries"] = 30
    minion_opts["grains"] = {}
    with patch("salt.loader.grains"):
        minion = salt.minion.Minion(minion_opts)

        load = {"load": "value"}
        timeout = 60

        # We are just validating no exception is raised
        rtn = await minion._send_req_async(load, timeout)
        assert rtn is False


def test_mine_send_tries(minion_opts):
    channel_enter = MagicMock()
    channel_enter.send.side_effect = lambda load, timeout, tries: tries
    channel = MagicMock()
    channel.__enter__.return_value = channel_enter

    minion_opts["return_retry_tries"] = 20
    with patch("salt.channel.client.ReqChannel.factory", return_value=channel), patch(
        "salt.loader.grains"
    ):
        minion = salt.minion.Minion(minion_opts)
        minion.tok = "token"

        data = {}
        tag = "tag"

        rtn = minion._mine_send(tag, data)
        assert rtn == 20


def test_invalid_master_address(minion_opts):
    minion_opts.update(
        {
            "ipv6": False,
            "master": float("127.0"),
            "master_port": "4555",
            "retry_dns": False,
        }
    )
    with pytest.raises(SaltSystemExit):
        salt.minion.resolve_dns(minion_opts)


def test_source_int_name_local(minion_opts):
    """
    test when file_client local and
    source_interface_name is set
    """
    interfaces = {
        "bond0.1234": {
            "hwaddr": "01:01:01:d0:d0:d0",
            "up": True,
            "inet": [
                {
                    "broadcast": "111.1.111.255",
                    "netmask": "111.1.0.0",
                    "label": "bond0",
                    "address": "111.1.0.1",
                }
            ],
        }
    }
    minion_opts.update(
        {
            "ipv6": False,
            "master": "127.0.0.1",
            "master_port": "4555",
            "file_client": "local",
            "source_interface_name": "bond0.1234",
            "source_ret_port": 49017,
            "source_publish_port": 49018,
        },
    )
    with patch("salt.utils.network.interfaces", MagicMock(return_value=interfaces)):
        assert salt.minion.resolve_dns(minion_opts) == {
            "master_ip": "127.0.0.1",
            "source_ip": "111.1.0.1",
            "source_ret_port": 49017,
            "source_publish_port": 49018,
            "master_uri": "tcp://127.0.0.1:4555",
        }


@pytest.mark.slow_test
def test_source_int_name_remote(minion_opts):
    """
    test when file_client remote and
    source_interface_name is set and
    interface is down
    """
    interfaces = {
        "bond0.1234": {
            "hwaddr": "01:01:01:d0:d0:d0",
            "up": False,
            "inet": [
                {
                    "broadcast": "111.1.111.255",
                    "netmask": "111.1.0.0",
                    "label": "bond0",
                    "address": "111.1.0.1",
                }
            ],
        }
    }
    minion_opts.update(
        {
            "ipv6": False,
            "master": "127.0.0.1",
            "master_port": "4555",
            "file_client": "remote",
            "source_interface_name": "bond0.1234",
            "source_ret_port": 49017,
            "source_publish_port": 49018,
        },
    )
    with patch("salt.utils.network.interfaces", MagicMock(return_value=interfaces)):
        assert salt.minion.resolve_dns(minion_opts) == {
            "master_ip": "127.0.0.1",
            "source_ret_port": 49017,
            "source_publish_port": 49018,
            "master_uri": "tcp://127.0.0.1:4555",
        }


@pytest.mark.slow_test
def test_source_address(minion_opts):
    """
    test when source_address is set
    """
    interfaces = {
        "bond0.1234": {
            "hwaddr": "01:01:01:d0:d0:d0",
            "up": False,
            "inet": [
                {
                    "broadcast": "111.1.111.255",
                    "netmask": "111.1.0.0",
                    "label": "bond0",
                    "address": "111.1.0.1",
                }
            ],
        }
    }
    minion_opts.update(
        {
            "ipv6": False,
            "master": "127.0.0.1",
            "master_port": "4555",
            "file_client": "local",
            "source_interface_name": "",
            "source_address": "111.1.0.1",
            "source_ret_port": 49017,
            "source_publish_port": 49018,
        },
    )
    with patch("salt.utils.network.interfaces", MagicMock(return_value=interfaces)):
        assert salt.minion.resolve_dns(minion_opts) == {
            "source_publish_port": 49018,
            "source_ret_port": 49017,
            "master_uri": "tcp://127.0.0.1:4555",
            "source_ip": "111.1.0.1",
            "master_ip": "127.0.0.1",
        }


# Tests for _handle_decoded_payload in the salt.minion.Minion() class: 3
@pytest.mark.slow_test
def test_handle_decoded_payload_jid_match_in_jid_queue(minion_opts):
    """
    Tests that the _handle_decoded_payload function returns when a jid is given that is already present
    in the jid_queue.

    Note: This test doesn't contain all of the patch decorators above the function like the other tests
    for _handle_decoded_payload below. This is essential to this test as the call to the function must
    return None BEFORE any of the processes are spun up because we should be avoiding firing duplicate
    jobs.
    """
    mock_data = {"fun": "foo.bar", "jid": 123}
    mock_jid_queue = [123]
    minion = salt.minion.Minion(
        minion_opts,
        jid_queue=copy.copy(mock_jid_queue),
        io_loop=salt.ext.tornado.ioloop.IOLoop(),
    )
    try:
        ret = minion._handle_decoded_payload(mock_data).result()
        assert minion.jid_queue == mock_jid_queue
        assert ret is None
    finally:
        minion.destroy()


@pytest.mark.slow_test
def test_handle_decoded_payload_jid_queue_addition(minion_opts):
    """
    Tests that the _handle_decoded_payload function adds a jid to the minion's jid_queue when the new
    jid isn't already present in the jid_queue.
    """
    with patch("salt.minion.Minion.ctx", MagicMock(return_value={})), patch(
        "salt.utils.process.SignalHandlingProcess.start",
        MagicMock(return_value=True),
    ), patch(
        "salt.utils.process.SignalHandlingProcess.join",
        MagicMock(return_value=True),
    ):
        mock_jid = 11111
        mock_data = {"fun": "foo.bar", "jid": mock_jid}
        mock_jid_queue = [123, 456]
        minion = salt.minion.Minion(
            minion_opts,
            jid_queue=copy.copy(mock_jid_queue),
            io_loop=salt.ext.tornado.ioloop.IOLoop(),
        )
        try:

            # Assert that the minion's jid_queue attribute matches the mock_jid_queue as a baseline
            # This can help debug any test failures if the _handle_decoded_payload call fails.
            assert minion.jid_queue == mock_jid_queue

            # Call the _handle_decoded_payload function and update the mock_jid_queue to include the new
            # mock_jid. The mock_jid should have been added to the jid_queue since the mock_jid wasn't
            # previously included. The minion's jid_queue attribute and the mock_jid_queue should be equal.
            minion._handle_decoded_payload(mock_data).result()
            mock_jid_queue.append(mock_jid)
            assert minion.jid_queue == mock_jid_queue
        finally:
            minion.destroy()


@pytest.mark.slow_test
def test_handle_decoded_payload_jid_queue_reduced_minion_jid_queue_hwm(minion_opts):
    """
    Tests that the _handle_decoded_payload function removes a jid from the minion's jid_queue when the
    minion's jid_queue high water mark (minion_jid_queue_hwm) is hit.
    """
    with patch("salt.minion.Minion.ctx", MagicMock(return_value={})), patch(
        "salt.utils.process.SignalHandlingProcess.start",
        MagicMock(return_value=True),
    ), patch(
        "salt.utils.process.SignalHandlingProcess.join",
        MagicMock(return_value=True),
    ):
        minion_opts["minion_jid_queue_hwm"] = 2
        mock_data = {"fun": "foo.bar", "jid": 789}
        mock_jid_queue = [123, 456]
        minion = salt.minion.Minion(
            minion_opts,
            jid_queue=copy.copy(mock_jid_queue),
            io_loop=salt.ext.tornado.ioloop.IOLoop(),
        )
        try:

            # Assert that the minion's jid_queue attribute matches the mock_jid_queue as a baseline
            # This can help debug any test failures if the _handle_decoded_payload call fails.
            assert minion.jid_queue == mock_jid_queue

            # Call the _handle_decoded_payload function and check that the queue is smaller by one item
            # and contains the new jid
            minion._handle_decoded_payload(mock_data).result()
            assert len(minion.jid_queue) == 2
            assert minion.jid_queue == [456, 789]
        finally:
            minion.destroy()


@pytest.mark.slow_test
def test_process_count_max(minion_opts):
    """
    Tests that the _handle_decoded_payload function does not spawn more than the configured amount of processes,
    as per process_count_max.
    """
    with patch("salt.minion.Minion.ctx", MagicMock(return_value={})), patch(
        "salt.utils.process.SignalHandlingProcess.start",
        MagicMock(return_value=True),
    ), patch(
        "salt.utils.process.SignalHandlingProcess.join",
        MagicMock(return_value=True),
    ), patch(
        "salt.utils.minion.running", MagicMock(return_value=[])
    ), patch(
        "salt.ext.tornado.gen.sleep",
        MagicMock(return_value=salt.ext.tornado.concurrent.Future()),
    ):
        process_count_max = 10
        minion_opts["__role"] = "minion"
        minion_opts["minion_jid_queue_hwm"] = 100
        minion_opts["process_count_max"] = process_count_max

        io_loop = salt.ext.tornado.ioloop.IOLoop()
        minion = salt.minion.Minion(minion_opts, jid_queue=[], io_loop=io_loop)
        try:

            # mock gen.sleep to throw a special Exception when called, so that we detect it
            class SleepCalledException(Exception):
                """Thrown when sleep is called"""

            salt.ext.tornado.gen.sleep.return_value.set_exception(
                SleepCalledException()
            )

            # up until process_count_max: gen.sleep does not get called, processes are started normally
            for i in range(process_count_max):
                mock_data = {"fun": "foo.bar", "jid": i}
                io_loop.run_sync(
                    lambda data=mock_data: minion._handle_decoded_payload(data)
                )
                assert (
                    salt.utils.process.SignalHandlingProcess.start.call_count == i + 1
                )
                assert len(minion.jid_queue) == i + 1
                salt.utils.minion.running.return_value += [i]

            # above process_count_max: gen.sleep does get called, JIDs are created but no new processes are started
            mock_data = {"fun": "foo.bar", "jid": process_count_max + 1}

            pytest.raises(
                SleepCalledException,
                lambda: io_loop.run_sync(
                    lambda: minion._handle_decoded_payload(mock_data)
                ),
            )
            assert (
                salt.utils.process.SignalHandlingProcess.start.call_count
                == process_count_max
            )
            assert len(minion.jid_queue) == process_count_max + 1
        finally:
            minion.destroy()


@pytest.mark.slow_test
def test_beacons_before_connect(minion_opts):
    """
    Tests that the 'beacons_before_connect' option causes the beacons to be initialized before connect.
    """
    with patch("salt.minion.Minion.ctx", MagicMock(return_value={})), patch(
        "salt.minion.Minion.sync_connect_master",
        MagicMock(side_effect=RuntimeError("stop execution")),
    ), patch(
        "salt.utils.process.SignalHandlingProcess.start",
        MagicMock(return_value=True),
    ), patch(
        "salt.utils.process.SignalHandlingProcess.join",
        MagicMock(return_value=True),
    ):
        minion_opts["beacons_before_connect"] = True
        io_loop = salt.ext.tornado.ioloop.IOLoop()
        io_loop.make_current()
        minion = salt.minion.Minion(minion_opts, io_loop=io_loop)
        try:

            try:
                minion.tune_in(start=True)
            except RuntimeError:
                pass

            # Make sure beacons are initialized but the sheduler is not
            assert "beacons" in minion.periodic_callbacks
            assert "schedule" not in minion.periodic_callbacks
        finally:
            minion.destroy()


@pytest.mark.slow_test
def test_scheduler_before_connect(minion_opts):
    """
    Tests that the 'scheduler_before_connect' option causes the scheduler to be initialized before connect.
    """
    with patch("salt.minion.Minion.ctx", MagicMock(return_value={})), patch(
        "salt.minion.Minion.sync_connect_master",
        MagicMock(side_effect=RuntimeError("stop execution")),
    ), patch(
        "salt.utils.process.SignalHandlingProcess.start",
        MagicMock(return_value=True),
    ), patch(
        "salt.utils.process.SignalHandlingProcess.join",
        MagicMock(return_value=True),
    ):
        minion_opts["scheduler_before_connect"] = True
        io_loop = salt.ext.tornado.ioloop.IOLoop()
        io_loop.make_current()
        minion = salt.minion.Minion(minion_opts, io_loop=io_loop)
        try:
            try:
                minion.tune_in(start=True)
            except RuntimeError:
                pass

            # Make sure the scheduler is initialized but the beacons are not
            assert "schedule" in minion.periodic_callbacks
            assert "beacons" not in minion.periodic_callbacks
        finally:
            minion.destroy()


def test_minion_module_refresh(minion_opts):
    """
    Tests that the 'module_refresh' just return in case there is no 'schedule'
    because destroy method was already called.
    """
    with patch("salt.minion.Minion.ctx", MagicMock(return_value={})), patch(
        "salt.utils.process.SignalHandlingProcess.start",
        MagicMock(return_value=True),
    ), patch(
        "salt.utils.process.SignalHandlingProcess.join",
        MagicMock(return_value=True),
    ):
        try:
            minion = salt.minion.Minion(
                minion_opts,
                io_loop=salt.ext.tornado.ioloop.IOLoop(),
            )
            minion.schedule = salt.utils.schedule.Schedule(
                minion_opts, {}, returners={}
            )
            assert hasattr(minion, "schedule")
            minion.destroy()
            assert not hasattr(minion, "schedule")
            assert not minion.module_refresh()
        finally:
            minion.destroy()


def test_minion_module_refresh_beacons_refresh(minion_opts):
    """
    Tests that 'module_refresh' calls beacons_refresh and that the
    minion object has a beacons attribute with beacons.
    """
    with patch("salt.minion.Minion.ctx", MagicMock(return_value={})), patch(
        "salt.utils.process.SignalHandlingProcess.start",
        MagicMock(return_value=True),
    ), patch(
        "salt.utils.process.SignalHandlingProcess.join",
        MagicMock(return_value=True),
    ):
        try:
            minion = salt.minion.Minion(
                minion_opts,
                io_loop=salt.ext.tornado.ioloop.IOLoop(),
            )
            minion.schedule = salt.utils.schedule.Schedule(
                minion_opts, {}, returners={}
            )
            assert not hasattr(minion, "beacons")
            minion.module_refresh()
            assert hasattr(minion, "beacons")
            assert hasattr(minion.beacons, "beacons")
            assert "service.beacon" in minion.beacons.beacons
            minion.destroy()
        finally:
            minion.destroy()


@pytest.mark.slow_test
def test_when_ping_interval_is_set_the_callback_should_be_added_to_periodic_callbacks(
    minion_opts,
):
    with patch("salt.minion.Minion.ctx", MagicMock(return_value={})), patch(
        "salt.minion.Minion.sync_connect_master",
        MagicMock(side_effect=RuntimeError("stop execution")),
    ), patch(
        "salt.utils.process.SignalHandlingProcess.start",
        MagicMock(return_value=True),
    ), patch(
        "salt.utils.process.SignalHandlingProcess.join",
        MagicMock(return_value=True),
    ):
        minion_opts["ping_interval"] = 10
        io_loop = salt.ext.tornado.ioloop.IOLoop()
        io_loop.make_current()
        minion = salt.minion.Minion(minion_opts, io_loop=io_loop)
        try:
            try:
                minion.connected = MagicMock(side_effect=(False, True))
                minion._fire_master_minion_start = MagicMock()
                minion.tune_in(start=False)
            except RuntimeError:
                pass

            # Make sure the scheduler is initialized but the beacons are not
            assert "ping" in minion.periodic_callbacks
        finally:
            minion.destroy()


@pytest.mark.slow_test
def test_when_passed_start_event_grains(minion_opts):
    # provide mock opts an os grain since we'll look for it later.
    minion_opts["grains"]["os"] = "linux"
    minion_opts["start_event_grains"] = ["os"]
    io_loop = salt.ext.tornado.ioloop.IOLoop()
    io_loop.make_current()
    minion = salt.minion.Minion(minion_opts, io_loop=io_loop)
    try:
        minion.tok = MagicMock()
        minion._send_req_sync = MagicMock()
        minion._fire_master(
            "Minion has started", "minion_start", include_startup_grains=True
        )
        load = minion._send_req_sync.call_args[0][0]

        assert "grains" in load
        assert "os" in load["grains"]
    finally:
        minion.destroy()


@pytest.mark.slow_test
def test_when_not_passed_start_event_grains(minion_opts):
    io_loop = salt.ext.tornado.ioloop.IOLoop()
    io_loop.make_current()
    minion = salt.minion.Minion(minion_opts, io_loop=io_loop)
    try:
        minion.tok = MagicMock()
        minion._send_req_sync = MagicMock()
        minion._fire_master("Minion has started", "minion_start")
        load = minion._send_req_sync.call_args[0][0]

        assert "grains" not in load
    finally:
        minion.destroy()


@pytest.mark.slow_test
def test_when_other_events_fired_and_start_event_grains_are_set(minion_opts):
    minion_opts["start_event_grains"] = ["os"]
    io_loop = salt.ext.tornado.ioloop.IOLoop()
    io_loop.make_current()
    minion = salt.minion.Minion(minion_opts, io_loop=io_loop)
    try:
        minion.tok = MagicMock()
        minion._send_req_sync = MagicMock()
        minion._fire_master("Custm_event_fired", "custom_event")
        load = minion._send_req_sync.call_args[0][0]

        assert "grains" not in load
    finally:
        minion.destroy()


@pytest.mark.slow_test
def test_minion_retry_dns_count(minion_opts):
    """
    Tests that the resolve_dns will retry dns look ups for a maximum of
    3 times before raising a SaltMasterUnresolvableError exception.
    """
    minion_opts.update(
        {
            "ipv6": False,
            "master": "dummy",
            "master_port": "4555",
            "retry_dns": 1,
            "retry_dns_count": 3,
        },
    )
    with pytest.raises(SaltMasterUnresolvableError):
        salt.minion.resolve_dns(minion_opts)


@pytest.mark.slow_test
def test_gen_modules_executors(minion_opts):
    """
    Ensure gen_modules is called with the correct arguments #54429
    """
    io_loop = salt.ext.tornado.ioloop.IOLoop()
    io_loop.make_current()
    minion = salt.minion.Minion(minion_opts, io_loop=io_loop)

    class MockPillarCompiler:
        def compile_pillar(self):
            return {}

    try:
        with patch("salt.pillar.get_pillar", return_value=MockPillarCompiler()):
            with patch("salt.loader.executors", mock=MagicMock()) as execmock:
                minion.gen_modules()
        execmock.assert_called_once_with(
            minion.opts, functions=minion.functions, proxy=minion.proxy, context={}
        )
    finally:
        minion.destroy()


def test_reinit_crypto_on_fork(minion_opts):
    """
    Ensure salt.utils.crypt.reinit_crypto() is executed when forking for new job
    """
    minion_opts["multiprocessing"] = True
    with patch("salt.utils.process.default_signals"):

        io_loop = salt.ext.tornado.ioloop.IOLoop()
        io_loop.make_current()
        minion = salt.minion.Minion(minion_opts, io_loop=io_loop)

        job_data = {"jid": "test-jid", "fun": "test.ping"}

        def mock_start(self):
            # pylint: disable=comparison-with-callable
            assert (
                len(
                    [
                        x
                        for x in self._after_fork_methods
                        if x[0] == salt.utils.crypt.reinit_crypto
                    ]
                )
                == 1
            )
            # pylint: enable=comparison-with-callable

        with patch.object(
            salt.utils.process.SignalHandlingProcess, "start", mock_start
        ):
            io_loop.run_sync(lambda: minion._handle_decoded_payload(job_data))


def test_minion_manage_schedule(minion_opts):
    """
    Tests that the manage_schedule will call the add function, adding
    schedule data into opts.
    """
    with patch("salt.minion.Minion.ctx", MagicMock(return_value={})), patch(
        "salt.minion.Minion.sync_connect_master",
        MagicMock(side_effect=RuntimeError("stop execution")),
    ), patch(
        "salt.utils.process.SignalHandlingProcess.start",
        MagicMock(return_value=True),
    ), patch(
        "salt.utils.process.SignalHandlingProcess.join",
        MagicMock(return_value=True),
    ):
        io_loop = salt.ext.tornado.ioloop.IOLoop()
        io_loop.make_current()

        with patch("salt.utils.schedule.clean_proc_dir", MagicMock(return_value=None)):
            try:
                mock_functions = {"test.ping": None}

                minion = salt.minion.Minion(minion_opts, io_loop=io_loop)
                minion.schedule = salt.utils.schedule.Schedule(
                    minion_opts,
                    mock_functions,
                    returners={},
                    new_instance=True,
                )

                minion.opts["foo"] = "bar"
                schedule_data = {
                    "test_job": {
                        "function": "test.ping",
                        "return_job": False,
                        "jid_include": True,
                        "maxrunning": 2,
                        "seconds": 10,
                    }
                }

                data = {
                    "name": "test-item",
                    "schedule": schedule_data,
                    "func": "add",
                    "persist": False,
                }
                tag = "manage_schedule"

                minion.manage_schedule(tag, data)
                assert "test_job" in minion.opts["schedule"]
            finally:
                del minion.schedule
                minion.destroy()
                del minion


def test_minion_manage_beacons(minion_opts):
    """
    Tests that the manage_beacons will call the add function, adding
    beacon data into opts.
    """
    with patch("salt.minion.Minion.ctx", MagicMock(return_value={})), patch(
        "salt.minion.Minion.sync_connect_master",
        MagicMock(side_effect=RuntimeError("stop execution")),
    ), patch(
        "salt.utils.process.SignalHandlingProcess.start",
        MagicMock(return_value=True),
    ), patch(
        "salt.utils.process.SignalHandlingProcess.join",
        MagicMock(return_value=True),
    ):
        try:
            minion_opts["beacons"] = {}

            io_loop = salt.ext.tornado.ioloop.IOLoop()
            io_loop.make_current()

            mock_functions = {"test.ping": None}
            minion = salt.minion.Minion(minion_opts, io_loop=io_loop)
            minion.beacons = salt.beacons.Beacon(minion_opts, mock_functions)

            bdata = [{"salt-master": "stopped"}, {"apache2": "stopped"}]
            data = {"name": "ps", "beacon_data": bdata, "func": "add"}

            tag = "manage_beacons"
            log.debug("==== minion.opts %s ====", minion.opts)

            minion.manage_beacons(tag, data)
            assert "ps" in minion.opts["beacons"]
            assert minion.opts["beacons"]["ps"] == bdata
        finally:
            minion.destroy()


def test_prep_ip_port():
    _ip = ipaddress.ip_address

    opts = {"master": "10.10.0.3", "master_uri_format": "ip_only"}
    ret = salt.minion.prep_ip_port(opts)
    assert ret == {"master": _ip("10.10.0.3")}

    opts = {
        "master": "10.10.0.3",
        "master_port": 1234,
        "master_uri_format": "default",
    }
    ret = salt.minion.prep_ip_port(opts)
    assert ret == {"master": "10.10.0.3"}

    opts = {"master": "10.10.0.3:1234", "master_uri_format": "default"}
    ret = salt.minion.prep_ip_port(opts)
    assert ret == {"master": "10.10.0.3", "master_port": 1234}

    opts = {"master": "host name", "master_uri_format": "default"}
    pytest.raises(SaltClientError, salt.minion.prep_ip_port, opts)

    opts = {"master": "10.10.0.3:abcd", "master_uri_format": "default"}
    pytest.raises(SaltClientError, salt.minion.prep_ip_port, opts)

    opts = {"master": "10.10.0.3::1234", "master_uri_format": "default"}
    pytest.raises(SaltClientError, salt.minion.prep_ip_port, opts)


@pytest.mark.skip_if_not_root
def test_sock_path_len(minion_opts):
    """
    This tests whether or not a larger hash causes the sock path to exceed
    the system's max sock path length. See the below link for more
    information.

    https://github.com/saltstack/salt/issues/12172#issuecomment-43903643
    """
    minion_opts.update(
        {
            "id": "salt-testing",
            "hash_type": "sha512",
            "sock_dir": os.path.join(salt.syspaths.SOCK_DIR, "minion"),
            "extension_modules": "",
        }
    )
    try:
        event_publisher = event.AsyncEventPublisher(minion_opts)
        result = True
    except ValueError:
        #  There are rare cases where we operate a closed socket, especially in containers.
        # In this case, don't fail the test because we'll catch it down the road.
        result = True
    except SaltSystemExit:
        result = False
    assert result


@pytest.mark.skip_on_windows(reason="Skippin, no Salt master running on Windows.")
async def test_master_type_failover(minion_opts):
    """
    Tests master_type "failover" to not fall back to 127.0.0.1 address when master does not resolve in DNS
    """
    minion_opts.update(
        {
            "master_type": "failover",
            "master": ["master1", "master2"],
            "__role": "",
            "retry_dns": 0,
        }
    )

    class MockPubChannel:
        def connect(self):
            raise SaltClientError("MockedChannel")

        def close(self):
            return

    def mock_resolve_dns(opts, fallback=False):
        assert not fallback

        if opts["master"] == "master1":
            raise SaltClientError("Cannot resolve {}".format(opts["master"]))

        return {
            "master_ip": "192.168.2.1",
            "master_uri": "tcp://192.168.2.1:4505",
        }

    def mock_channel_factory(opts, **kwargs):
        assert opts["master"] == "master2"
        return MockPubChannel()

    with patch("salt.minion.resolve_dns", mock_resolve_dns), patch(
        "salt.channel.client.AsyncPubChannel.factory", mock_channel_factory
    ), patch("salt.loader.grains", MagicMock(return_value=[])):
        with pytest.raises(SaltClientError):
            minion = salt.minion.Minion(minion_opts)
            await minion.connect_master()


async def test_master_type_failover_no_masters(minion_opts):
    """
    Tests master_type "failover" to not fall back to 127.0.0.1 address when no master can be resolved
    """
    minion_opts.update(
        {
            "master_type": "failover",
            "master": ["master1", "master2"],
            "__role": "",
            "retry_dns": 0,
        }
    )

    def mock_resolve_dns(opts, fallback=False):
        assert not fallback
        raise SaltClientError("Cannot resolve {}".format(opts["master"]))

    with patch("salt.minion.resolve_dns", mock_resolve_dns), patch(
        "salt.loader.grains", MagicMock(return_value=[])
    ):
        with pytest.raises(SaltClientError):
            minion = salt.minion.Minion(minion_opts)
            await minion.connect_master()


def test_config_cache_path_overrides():
    cachedir = os.path.abspath("/path/to/master/cache")
    opts = {"cachedir": cachedir, "conf_file": None}

    mminion = salt.minion.MasterMinion(opts)
    assert mminion.opts["cachedir"] == cachedir


def test_minion_grains_refresh_pre_exec_false(minion_opts):
    """
    Minion does not refresh grains when grains_refresh_pre_exec is False
    """
    minion_opts["multiprocessing"] = False
    minion_opts["grains_refresh_pre_exec"] = False
    mock_data = {"fun": "foo.bar", "jid": 123}
    with patch("salt.loader.grains") as grainsfunc, patch(
        "salt.minion.Minion._target", MagicMock(return_value=True)
    ):
        minion = salt.minion.Minion(
            minion_opts,
            jid_queue=None,
            io_loop=salt.ext.tornado.ioloop.IOLoop(),
            load_grains=False,
        )
        try:
            ret = minion._handle_decoded_payload(mock_data).result()
            grainsfunc.assert_not_called()
        finally:
            minion.destroy()


def test_minion_grains_refresh_pre_exec_true(minion_opts):
    """
    Minion refreshes grains when grains_refresh_pre_exec is True
    """
    minion_opts["multiprocessing"] = False
    minion_opts["grains_refresh_pre_exec"] = True
    mock_data = {"fun": "foo.bar", "jid": 123}
    with patch("salt.loader.grains") as grainsfunc, patch(
        "salt.minion.Minion._target", MagicMock(return_value=True)
    ):
        minion = salt.minion.Minion(
            minion_opts,
            jid_queue=None,
            io_loop=salt.ext.tornado.ioloop.IOLoop(),
            load_grains=False,
        )
        try:
            ret = minion._handle_decoded_payload(mock_data).result()
            grainsfunc.assert_called()
        finally:
            minion.destroy()


@pytest.mark.skip_on_darwin(
    reason="Skip on MacOS, where this does not raise an exception."
)
def test_valid_ipv4_master_address_ipv6_enabled(minion_opts):
    """
    Tests that the lookups fail back to ipv4 when ipv6 fails.
    """
    interfaces = {
        "bond0.1234": {
            "hwaddr": "01:01:01:d0:d0:d0",
            "up": False,
            "inet": [
                {
                    "broadcast": "111.1.111.255",
                    "netmask": "111.1.0.0",
                    "label": "bond0",
                    "address": "111.1.0.1",
                }
            ],
        }
    }
    minion_opts.update(
        {
            "ipv6": True,
            "master": "127.0.0.1",
            "master_port": "4555",
            "retry_dns": False,
            "source_address": "111.1.0.1",
            "source_interface_name": "bond0.1234",
            "source_ret_port": 49017,
            "source_publish_port": 49018,
        },
    )
    with patch("salt.utils.network.interfaces", MagicMock(return_value=interfaces)):
        expected = {
            "source_publish_port": 49018,
            "master_uri": "tcp://127.0.0.1:4555",
            "source_ret_port": 49017,
            "master_ip": "127.0.0.1",
        }
        assert salt.minion.resolve_dns(minion_opts) == expected


async def test_master_type_disable(minion_opts):
    """
    Tests master_type "disable" to not even attempt connecting to a master.
    """
    minion_opts.update(
        {
            "master_type": "disable",
            "master": None,
            "__role": "",
            "pub_ret": False,
            "file_client": "local",
        }
    )

    minion = salt.minion.Minion(minion_opts)
    try:

        try:
            minion_man = salt.minion.MinionManager(minion_opts)
            minion_man._connect_minion(minion)
        except RuntimeError:
            pytest.fail("_connect_minion(minion) threw an error, This was not expected")

        # Make sure beacons and sheduler are initialized
        assert "beacons" in minion.periodic_callbacks
        assert "schedule" in minion.periodic_callbacks
        assert minion.connected is False
    finally:
        minion.destroy()


async def test_syndic_async_req_channel(syndic_opts):
    syndic_opts["_minion_conf_file"] = ""
    syndic_opts["master_uri"] = "tcp://127.0.0.1:4506"
    syndic = salt.minion.Syndic(syndic_opts)
    syndic.pub_channel = MagicMock()
    syndic.tune_in_no_block()
    assert isinstance(syndic.async_req_channel, salt.channel.client.AsyncReqChannel)



"""
    :codeauthor: Rupesh Tare <rupesht@saltstack.com>
    :codeauthor: Herbert Buurman <herbert.buurman@ogd.nl>
"""


import pytest

import salt.modules.mine as mine
import salt.utils.mine
from salt.utils.odict import OrderedDict
from tests.support.mock import MagicMock, patch


class FakeCache:
    def __init__(self):
        self.data = {}

    def store(self, bank, key, value):
        self.data[bank, key] = value
        return "FakeCache:StoreSuccess!"

    def fetch(self, bank, key):
        return self.data.get((bank, key), {})

    def debug(self):
        print("{}:FakeCache dump:\n{}".format(__name__, self.data))


@pytest.fixture
def mock_cache():
    cache = FakeCache()
    return cache


@pytest.fixture
def configure_loader_modules(mock_cache):
    mock_match = MagicMock(return_value="webserver")
    return {
        mine: {
            "__salt__": {
                "match.glob": mock_match,
                "match.pcre": mock_match,
                "match.list": mock_match,
                "match.grain": mock_match,
                "match.grain_pcre": mock_match,
                "match.ipcidr": mock_match,
                "match.compound": mock_match,
                "match.pillar": mock_match,
                "match.pillar_pcre": mock_match,
                "data.get": lambda key: mock_cache.fetch("minions/webserver", key),
                "data.update": lambda key, value: mock_cache.store(
                    "minions/webserver", key, value
                ),
            }
        }
    }


def test_get_local_empty():
    """
    Tests getting function data from the local mine that does not exist.
    """
    with patch.dict(mine.__opts__, {"file_client": "local", "id": "webserver"}):
        ret_classic = mine.get("*", "funky.doodle")
        ret_dict = mine.get("*", ["funky.doodle"])
    assert ret_classic == {}
    assert ret_dict == {}


def test_get_local_classic(mock_cache):
    """
    Tests getting function data from the local mine that was stored without minion-side ACL.
    This verifies backwards compatible reads from a salt mine.
    """
    # Prefill minion cache with a non-ACL value
    mock_cache.store("minions/webserver", "mine_cache", {"foobard": "barfood"})
    with patch.dict(mine.__opts__, {"file_client": "local", "id": "webserver"}):
        ret_classic = mine.get("*", "foobard")
        ret_dict = mine.get("*", ["foobard"])
    assert ret_classic == {"webserver": "barfood"}
    assert ret_dict == {"foobard": {"webserver": "barfood"}}


def test_send_get_local(mock_cache):
    """
    Tests sending an item to the mine in the minion's local cache,
    and then immediately fetching it again (since tests are executed unordered).
    Also verify that the stored mine cache does not use ACL data structure
    without allow_tgt passed.
    """
    foo_ret = "baz"
    ip_ret = "2001:db8::1:3"
    with patch.dict(
        mine.__opts__, {"file_client": "local", "id": "webserver"}
    ), patch.dict(
        mine.__salt__,
        {
            "network.ip_addrs": MagicMock(return_value=ip_ret),
            "foo.bar": MagicMock(return_value=foo_ret),
        },
    ):
        ret = mine.send("ip_addr", mine_function="network.ip_addrs")
        mine.send("foo.bar")
    assert ret == "FakeCache:StoreSuccess!"
    assert mock_cache.fetch("minions/webserver", "mine_cache") == {
        "ip_addr": ip_ret,
        "foo.bar": foo_ret,
    }
    with patch.dict(mine.__opts__, {"file_client": "local", "id": "webserver"}):
        ret_single = mine.get("*", "ip_addr")
        ret_single_dict = mine.get("*", ["ip_addr"])
        ret_multi = mine.get("*", "ip_addr,foo.bar")
        ret_multi2 = mine.get("*", ["ip_addr", "foo.bar"])
    assert ret_single == {"webserver": ip_ret}
    assert ret_single_dict == {"ip_addr": {"webserver": ip_ret}}
    assert ret_multi == {
        "ip_addr": {"webserver": ip_ret},
        "foo.bar": {"webserver": foo_ret},
    }
    assert ret_multi == ret_multi2


def test_send_get_acl_local(mock_cache):
    """
    Tests sending an item to the mine in the minion's local cache,
    including ACL information (useless when only working locally, but hey),
    and then immediately fetching it again (since tests are executed unordered).
    Also verify that the stored mine cache has the correct structure (with ACL)
    when using allow_tgt and no ACL without allow_tgt.
    """
    foo_ret = "baz"
    ip_ret = "2001:db8::1:3"
    with patch.dict(
        mine.__opts__, {"file_client": "local", "id": "webserver"}
    ), patch.dict(
        mine.__salt__,
        {
            "network.ip_addrs": MagicMock(return_value=ip_ret),
            "foo.bar": MagicMock(return_value=foo_ret),
        },
    ):
        ret = mine.send(
            "ip_addr",
            mine_function="network.ip_addrs",
            allow_tgt="web*",
            allow_tgt_type="glob",
        )
        mine.send("foo.bar")
    assert ret == "FakeCache:StoreSuccess!"
    assert mock_cache.fetch("minions/webserver", "mine_cache") == {
        "ip_addr": {
            salt.utils.mine.MINE_ITEM_ACL_DATA: ip_ret,
            salt.utils.mine.MINE_ITEM_ACL_ID: salt.utils.mine.MINE_ITEM_ACL_VERSION,
            "allow_tgt": "web*",
            "allow_tgt_type": "glob",
        },
        "foo.bar": foo_ret,
    }
    with patch.dict(mine.__opts__, {"file_client": "local", "id": "webserver"}):
        ret_single = mine.get("*", "ip_addr")
    assert ret_single == {"webserver": ip_ret}


def test_send_master():
    """
    Tests sending an item to the mine stored on the master.
    This is done by capturing the load that is sent to the master.
    """
    foo_ret = "baz"
    with patch.object(
        mine, "_mine_send", MagicMock(side_effect=lambda x, y: x)
    ), patch.dict(
        mine.__salt__, {"foo.bar": MagicMock(return_value=foo_ret)}
    ), patch.dict(
        mine.__opts__, {"file_client": "remote", "id": "foo"}
    ):
        ret = mine.send("foo.bar")
    assert ret == {
        "id": "foo",
        "cmd": "_mine",
        "data": {"foo.bar": foo_ret},
        "clear": False,
    }


def test_send_master_acl():
    """
    Tests sending an item to the mine stored on the master. Now with ACL.
    This is done by capturing the load that is sent to the master.
    """
    foo_ret = "baz"
    with patch.object(
        mine, "_mine_send", MagicMock(side_effect=lambda x, y: x)
    ), patch.dict(
        mine.__salt__, {"foo.bar": MagicMock(return_value=foo_ret)}
    ), patch.dict(
        mine.__opts__, {"file_client": "remote", "id": "foo"}
    ):
        ret = mine.send("foo.bar", allow_tgt="roles:web", allow_tgt_type="grains")
    assert ret == {
        "id": "foo",
        "cmd": "_mine",
        "data": {
            "foo.bar": {
                salt.utils.mine.MINE_ITEM_ACL_DATA: foo_ret,
                salt.utils.mine.MINE_ITEM_ACL_ID: salt.utils.mine.MINE_ITEM_ACL_VERSION,
                "allow_tgt": "roles:web",
                "allow_tgt_type": "grains",
            },
        },
        "clear": False,
    }


def test_get_master():
    """
    Tests loading a mine item from the mine stored on the master.
    """
    foo_ret = "baz"
    mock_load = {
        "tgt_type": "qux",
        "tgt": foo_ret,
        "cmd": "_mine_get",
        "fun": "foo.bar",
        "id": "foo",
    }
    with patch.object(mine, "_mine_get", MagicMock(return_value=mock_load)), patch.dict(
        mine.__opts__, {"file_client": "remote", "id": "foo"}
    ):
        assert mine.get("*", "foo.bar") == mock_load


def test_get_master_exclude_minion():
    """
    Tests the exclude_minion-parameter for mine.get
    """
    _mine_get_ret = OrderedDict([("webserver", "value")])
    with patch.object(
        mine, "_mine_get", MagicMock(return_value=_mine_get_ret)
    ), patch.dict(mine.__opts__, {"file_client": "remote", "id": "webserver"}):
        assert mine.get("*", "foo.bar", exclude_minion=False) == {"webserver": "value"}
        assert mine.get("*", "foo.bar", exclude_minion=True) == {}


def test_update_local(mock_cache):
    """
    Tests the ``update``-function on the minion's local cache.
    Updates mine functions from pillar+config only.
    """
    kernel_ret = "Linux!"
    foo_ret = "baz"
    ip_ret = "2001:db8::1:3"
    config_mine_functions = {
        "ip_addr": {"mine_function": "network.ip_addrs"},
        "network.ip_addrs": [],
        "kernel": [
            {"mine_function": "grains.get"},
            "kernel",
            {"allow_tgt": "web*"},
        ],
        "foo.bar": {"allow_tgt": "G@roles:webserver", "allow_tgt_type": "compound"},
    }
    with patch.dict(
        mine.__opts__, {"file_client": "local", "id": "webserver"}
    ), patch.dict(
        mine.__salt__,
        {
            "config.merge": MagicMock(return_value=config_mine_functions),
            "grains.get": lambda x: kernel_ret,
            "network.ip_addrs": MagicMock(return_value=ip_ret),
            "foo.bar": MagicMock(return_value=foo_ret),
        },
    ):
        ret = mine.update()
    assert ret == "FakeCache:StoreSuccess!"
    # Check if the mine entries have been stored properly in the FakeCache.
    assert mock_cache.fetch("minions/webserver", "mine_cache") == {
        "ip_addr": ip_ret,
        "network.ip_addrs": ip_ret,
        "foo.bar": {
            salt.utils.mine.MINE_ITEM_ACL_DATA: foo_ret,
            salt.utils.mine.MINE_ITEM_ACL_ID: salt.utils.mine.MINE_ITEM_ACL_VERSION,
            "allow_tgt": "G@roles:webserver",
            "allow_tgt_type": "compound",
        },
        "kernel": {
            salt.utils.mine.MINE_ITEM_ACL_DATA: kernel_ret,
            salt.utils.mine.MINE_ITEM_ACL_ID: salt.utils.mine.MINE_ITEM_ACL_VERSION,
            "allow_tgt": "web*",
        },
    }


def test_update_local_specific(mock_cache):
    """
    Tests the ``update``-function on the minion's local cache.
    Updates mine functions from kwargs only.
    """
    foo_ret = "baz"
    ip_ret = "2001:db8::1:3"
    manual_mine_functions = {
        "ip_addr": {"mine_function": "network.ip_addrs"},
        "network.ip_addrs": [],
        "kernel": [
            {"mine_function": "grains.get"},
            "kernel",
            {"allow_tgt": "web*"},
        ],
        "foo.bar": {"allow_tgt": "G@roles:webserver", "allow_tgt_type": "compound"},
    }
    with patch.dict(
        mine.__opts__, {"file_client": "local", "id": "webserver"}
    ), patch.dict(
        mine.__salt__,
        {
            "config.merge": MagicMock(return_value={}),
            "grains.get": lambda x: "Linux!!",
            "network.ip_addrs": MagicMock(return_value=ip_ret),
            "foo.bar": MagicMock(return_value=foo_ret),
        },
    ):
        ret = mine.update(mine_functions=manual_mine_functions)
    assert ret == "FakeCache:StoreSuccess!"
    assert mock_cache.fetch("minions/webserver", "mine_cache") == {
        "ip_addr": ip_ret,
        "network.ip_addrs": ip_ret,
        "foo.bar": {
            salt.utils.mine.MINE_ITEM_ACL_DATA: foo_ret,
            salt.utils.mine.MINE_ITEM_ACL_ID: salt.utils.mine.MINE_ITEM_ACL_VERSION,
            "allow_tgt": "G@roles:webserver",
            "allow_tgt_type": "compound",
        },
        "kernel": {
            salt.utils.mine.MINE_ITEM_ACL_DATA: "Linux!!",
            salt.utils.mine.MINE_ITEM_ACL_ID: salt.utils.mine.MINE_ITEM_ACL_VERSION,
            "allow_tgt": "web*",
        },
    }


def test_update_master():
    """
    Tests whether the ``update``-function sends the correct data to the master.
    """
    kernel_ret = "Linux!"
    foo_ret = "baz"
    ip_ret = "2001:db8::1:3"
    config_mine_functions = {
        "ip_addr": {"mine_function": "network.ip_addrs"},
        "network.ip_addrs": [],
        "kernel": [{"mine_function": "grains.get"}, "kernel"],
        "foo.bar": {},
    }
    mock_load = {
        "id": "webserver",
        "cmd": "_mine",
        "data": {
            "ip_addr": ip_ret,
            "network.ip_addrs": ip_ret,
            "foo.bar": foo_ret,
            "kernel": kernel_ret,
        },
        "clear": False,
    }
    with patch.object(
        mine, "_mine_send", MagicMock(side_effect=lambda x, y: x)
    ), patch.dict(
        mine.__opts__, {"file_client": "remote", "id": "webserver"}
    ), patch.dict(
        mine.__salt__,
        {
            "config.merge": MagicMock(return_value=config_mine_functions),
            "grains.get": lambda x: kernel_ret,
            "network.ip_addrs": MagicMock(return_value=ip_ret),
            "foo.bar": MagicMock(return_value=foo_ret),
        },
    ):
        # Verify the correct load
        assert mine.update() == mock_load


def test_delete_local(mock_cache):
    """
    Tests the ``delete``-function on the minion's local cache.
    """
    # Prefill minion cache with a non-ACL value
    mock_cache.store("minions/webserver", "mine_cache", {"foobard": "barfood"})
    with patch.dict(mine.__opts__, {"file_client": "local", "id": "webserver"}):
        ret = mine.delete("foobard")
        assert mock_cache.fetch("minions/webserver", "mine_cache") == {}


def test_delete_master(mock_cache):
    """
    Tests whether the ``delete``-function sends the correct data to the master.
    """
    # Prefill minion cache with a non-ACL value
    mock_cache.store("minions/webserver", "mine_cache", {"foobard": "barfood"})
    mock_load = {
        "cmd": "_mine_delete",
        "fun": "foobard",
        "id": "foo",
    }
    with patch.object(
        mine, "_mine_send", MagicMock(side_effect=lambda x, y: x)
    ), patch.dict(mine.__opts__, {"file_client": "remote", "id": "foo"}):
        assert mine.delete("foobard") == mock_load


def test_flush_local(mock_cache):
    """
    Tests the ``flush``-function on the minion's local cache.
    """
    # Prefill minion cache with a non-ACL value
    mock_cache.store("minions/webserver", "mine_cache", {"foobard": "barfood"})
    with patch.dict(mine.__opts__, {"file_client": "local", "id": "webserver"}):
        ret = mine.flush()
        assert mock_cache.fetch("minions/webserver", "mine_cache") == {}


def test_flush_master():
    """
    Tests whether the ``flush``-function sends the correct data to the master.
    """
    mock_load = {"cmd": "_mine_flush", "id": "foo"}
    with patch.object(
        mine, "_mine_send", MagicMock(side_effect=lambda x, y: x)
    ), patch.dict(mine.__opts__, {"file_client": "remote", "id": "foo"}):
        assert mine.flush() == mock_load


def test_valid():
    """
    Tests the ``valid``-function.
    Note that mine functions defined as list are returned in dict format.
    Mine functions that do not exist in __salt__ are not returned.
    """
    config_mine_functions = {
        "network.ip_addrs": [],
        "kernel": [
            {"mine_function": "grains.get"},
            "kernel",
            {"os": "win32", "v": "2018"},
        ],
        "fubar": [{"mine_function": "does.not_exist"}],
    }
    with patch.dict(
        mine.__salt__,
        {
            "config.merge": MagicMock(return_value=config_mine_functions),
            "network.ip_addrs": lambda: True,
            "grains.get": lambda: True,
        },
    ):
        ret = mine.valid()
        # list cant be made to set "dict can't be hashed" and order changes
        assert isinstance(ret["kernel"]["grains.get"], list)
        assert len(ret["kernel"]["grains.get"]) == 3
        for item in ("kernel", {"os": "win32"}, {"v": "2018"}):
            assert item in ret["kernel"]["grains.get"]
        ret["kernel"]["grains.get"] = None

        assert ret == {"network.ip_addrs": [], "kernel": {"grains.get": None}}


def test_get_docker():
    """
    Test for Get all mine data for 'docker.ps' and run an
    aggregation.
    """
    ps_response = {
        "localhost": {
            "host": {
                "interfaces": {
                    "docker0": {
                        "hwaddr": "88:99:00:00:99:99",
                        "inet": [
                            {
                                "address": "172.17.42.1",
                                "broadcast": None,
                                "label": "docker0",
                                "netmask": "255.255.0.0",
                            }
                        ],
                        "inet6": [
                            {
                                "address": "ffff::eeee:aaaa:bbbb:8888",
                                "prefixlen": "64",
                            }
                        ],
                        "up": True,
                    },
                    "eth0": {
                        "hwaddr": "88:99:00:99:99:99",
                        "inet": [
                            {
                                "address": "192.168.0.1",
                                "broadcast": "192.168.0.255",
                                "label": "eth0",
                                "netmask": "255.255.255.0",
                            }
                        ],
                        "inet6": [
                            {
                                "address": "ffff::aaaa:aaaa:bbbb:8888",
                                "prefixlen": "64",
                            }
                        ],
                        "up": True,
                    },
                }
            },
            "abcdefhjhi1234567899": {  # container Id
                "Ports": [
                    {
                        "IP": "0.0.0.0",  # we bind on every interfaces
                        "PrivatePort": 80,
                        "PublicPort": 80,
                        "Type": "tcp",
                    }
                ],
                "Image": "image:latest",
                "Info": {"Id": "abcdefhjhi1234567899"},
            },
        }
    }
    with patch.object(mine, "get", return_value=ps_response):
        ret = mine.get_docker()
        # Sort ifaces since that will change between py2 and py3
        ret["image:latest"]["ipv4"][80] = sorted(ret["image:latest"]["ipv4"][80])
        assert ret == {
            "image:latest": {"ipv4": {80: sorted(["172.17.42.1:80", "192.168.0.1:80"])}}
        }


def test_get_docker_with_container_id():
    """
    Test for Get all mine data for 'docker.ps' and run an
    aggregation.
    """
    ps_response = {
        "localhost": {
            "host": {
                "interfaces": {
                    "docker0": {
                        "hwaddr": "88:99:00:00:99:99",
                        "inet": [
                            {
                                "address": "172.17.42.1",
                                "broadcast": None,
                                "label": "docker0",
                                "netmask": "255.255.0.0",
                            }
                        ],
                        "inet6": [
                            {
                                "address": "ffff::eeee:aaaa:bbbb:8888",
                                "prefixlen": "64",
                            }
                        ],
                        "up": True,
                    },
                    "eth0": {
                        "hwaddr": "88:99:00:99:99:99",
                        "inet": [
                            {
                                "address": "192.168.0.1",
                                "broadcast": "192.168.0.255",
                                "label": "eth0",
                                "netmask": "255.255.255.0",
                            }
                        ],
                        "inet6": [
                            {
                                "address": "ffff::aaaa:aaaa:bbbb:8888",
                                "prefixlen": "64",
                            }
                        ],
                        "up": True,
                    },
                }
            },
            "abcdefhjhi1234567899": {  # container Id
                "Ports": [
                    {
                        "IP": "0.0.0.0",  # we bind on every interfaces
                        "PrivatePort": 80,
                        "PublicPort": 80,
                        "Type": "tcp",
                    }
                ],
                "Image": "image:latest",
                "Info": {"Id": "abcdefhjhi1234567899"},
            },
        }
    }
    with patch.object(mine, "get", return_value=ps_response):
        ret = mine.get_docker(with_container_id=True)
        # Sort ifaces since that will change between py2 and py3
        ret["image:latest"]["ipv4"][80] = sorted(ret["image:latest"]["ipv4"][80])
        assert ret == {
            "image:latest": {
                "ipv4": {
                    80: sorted(
                        [
                            ("172.17.42.1:80", "abcdefhjhi1234567899"),
                            ("192.168.0.1:80", "abcdefhjhi1234567899"),
                        ]
                    )
                }
            }
        }



"""
    Unit test for salt.grains.metadata_azure


    :codeauthor: :email" `Vishal Gupta <guvishal@vmware.com>

"""

import logging

import pytest

import salt.grains.metadata_azure as metadata
import salt.utils.http as http
from tests.support.mock import create_autospec, patch

# from Exception import Exception, ValueError

log = logging.getLogger(__name__)


@pytest.fixture
def configure_loader_modules():
    return {metadata: {"__opts__": {"metadata_server_grains": "True"}}}


def test_metadata_azure_search():
    def mock_http(url="", headers=False, header_list=None):
        metadata_vals = {
            "http://169.254.169.254/metadata/instance?api-version=2020-09-01": {
                "body": '{"compute": {"test": "fulltest"}}',
                "headers": {"Content-Type": "application/json; charset=utf-8"},
            },
        }

        return metadata_vals[url]

    with patch(
        "salt.utils.http.query",
        create_autospec(http.query, autospec=True, side_effect=mock_http),
    ):
        assert metadata.metadata() == {"compute": {"test": "fulltest"}}


def test_metadata_virtual():
    print("running 1st")
    with patch(
        "salt.utils.http.query",
        create_autospec(
            http.query,
            autospec=True,
            return_value={
                "error": "Bad request: . Required metadata header not specified"
            },
        ),
    ):
        assert metadata.__virtual__() is False
    with patch(
        "salt.utils.http.query",
        create_autospec(
            http.query,
            autospec=True,
            return_value={
                "body": '{"compute": {"test": "fulltest"}}',
                "headers": {"Content-Type": "application/json; charset=utf-8"},
                "status": 200,
            },
        ),
    ):
        assert metadata.__virtual__() is True
    with patch(
        "salt.utils.http.query",
        create_autospec(
            http.query,
            autospec=True,
            return_value={
                "body": "test",
                "headers": {"Content-Type": "application/json; charset=utf-8"},
                "status": 404,
            },
        ),
    ):
        assert metadata.__virtual__() is False
    with patch(
        "salt.utils.http.query",
        create_autospec(
            http.query,
            autospec=True,
            return_value={
                "body": "test",
                "headers": {"Content-Type": "application/json; charset=utf-8"},
                "status": 400,
            },
        ),
    ):
        assert metadata.__virtual__() is False



"""
Beacon to emit Telegram messages

Requires the python-telegram-bot library

"""
import logging

import salt.utils.beacons

try:
    import telegram

    logging.getLogger("telegram").setLevel(logging.CRITICAL)
    HAS_TELEGRAM = True
except ImportError:
    HAS_TELEGRAM = False

log = logging.getLogger(__name__)


__virtualname__ = "telegram_bot_msg"


def __virtual__():
    if HAS_TELEGRAM:
        return __virtualname__
    else:
        err_msg = "telegram library is missing."
        log.error("Unable to load %s beacon: %s", __virtualname__, err_msg)
        return False, err_msg


def validate(config):
    """
    Validate the beacon configuration
    """
    if not isinstance(config, list):
        return False, "Configuration for telegram_bot_msg beacon must be a list."

    config = salt.utils.beacons.list_to_dict(config)

    if not all(
        config.get(required_config) for required_config in ["token", "accept_from"]
    ):
        return (
            False,
            "Not all required configuration for telegram_bot_msg are set.",
        )

    if not isinstance(config.get("accept_from"), list):
        return (
            False,
            "Configuration for telegram_bot_msg, "
            "accept_from must be a list of usernames.",
        )

    return True, "Valid beacon configuration."


def beacon(config):
    """
    Emit a dict with a key "msgs" whose value is a list of messages
    sent to the configured bot by one of the allowed usernames.

    .. code-block:: yaml

        beacons:
          telegram_bot_msg:
            - token: "<bot access token>"
            - accept_from:
              - "<valid username>"
            - interval: 10

    """

    config = salt.utils.beacons.list_to_dict(config)

    log.debug("telegram_bot_msg beacon starting")
    ret = []
    output = {}
    output["msgs"] = []

    bot = telegram.Bot(config["token"])
    updates = bot.get_updates(limit=100, timeout=0)

    log.debug("Num updates: %d", len(updates))
    if not updates:
        log.debug("Telegram Bot beacon has no new messages")
        return ret

    latest_update_id = 0
    for update in updates:
        message = update.message

        if update.update_id > latest_update_id:
            latest_update_id = update.update_id

        if message.chat.username in config["accept_from"]:
            output["msgs"].append(message.to_dict())

    # mark in the server that previous messages are processed
    bot.get_updates(offset=latest_update_id + 1)

    log.debug("Emitting %d messages.", len(output["msgs"]))
    if output["msgs"]:
        ret.append(output)
    return ret



import logging
import os
import random
import string
from tempfile import NamedTemporaryFile

import pytest

import salt.config
import salt.loader
import salt.modules.boto_lambda as boto_lambda
import salt.utils.stringutils
from salt.exceptions import SaltInvocationError
from salt.utils.versions import Version
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase

# pylint: disable=import-error,no-name-in-module
try:
    import boto3
    from botocore import __version__ as found_botocore_version
    from botocore.exceptions import ClientError

    HAS_BOTO = True
except ImportError:
    HAS_BOTO = False

# pylint: enable=import-error,no-name-in-module

# the boto_lambda module relies on the connect_to_region() method
# which was added in boto 2.8.0
# https://github.com/boto/boto/commit/33ac26b416fbb48a60602542b4ce15dcc7029f12
required_boto3_version = "1.2.1"
required_botocore_version = "1.5.2"

region = "us-east-1"
access_key = "GKTADJGHEIQSXMKKRBJ08H"
secret_key = "askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs"
conn_parameters = {
    "region": region,
    "key": access_key,
    "keyid": secret_key,
    "profile": {},
}
error_message = (
    "An error occurred (101) when calling the {0} operation: Test-defined error"
)
error_content = {"Error": {"Code": 101, "Message": "Test-defined error"}}
function_ret = dict(
    FunctionName="testfunction",
    Runtime="python2.7",
    Role=None,
    Handler="handler",
    Description="abcdefg",
    Timeout=5,
    MemorySize=128,
    CodeSha256="abcdef",
    CodeSize=199,
    FunctionArn="arn:lambda:us-east-1:1234:Something",
    LastModified="yes",
    VpcConfig=None,
    Environment=None,
)
alias_ret = dict(
    AliasArn="arn:lambda:us-east-1:1234:Something",
    Name="testalias",
    FunctionVersion="3",
    Description="Alias description",
)
event_source_mapping_ret = dict(
    UUID="1234-1-123",
    BatchSize=123,
    EventSourceArn="arn:lambda:us-east-1:1234:Something",
    FunctionArn="arn:lambda:us-east-1:1234:Something",
    LastModified="yes",
    LastProcessingResult="SUCCESS",
    State="Enabled",
    StateTransitionReason="Random",
)

log = logging.getLogger(__name__)


def _has_required_boto():
    """
    Returns True/False boolean depending on if Boto is installed and correct
    version.
    """
    if not HAS_BOTO:
        return False
    elif Version(boto3.__version__) < Version(required_boto3_version):
        return False
    elif Version(found_botocore_version) < Version(required_botocore_version):
        return False
    else:
        return True


@pytest.mark.skipif(HAS_BOTO is False, reason="The boto module must be installed.")
@pytest.mark.skipif(
    _has_required_boto() is False,
    reason="The boto3 module must be greater than or equal to version {}, "
    "and botocore must be greater than or equal to {}".format(
        required_boto3_version, required_botocore_version
    ),
)
class BotoLambdaTestCaseBase(TestCase, LoaderModuleMockMixin):
    conn = None

    def setup_loader_modules(self):
        self.opts = opts = salt.config.DEFAULT_MINION_OPTS.copy()
        utils = salt.loader.utils(
            opts, whitelist=["boto3", "args", "systemd", "path", "platform"], context={}
        )
        return {boto_lambda: {"__utils__": utils}}

    def setUp(self):
        super().setUp()
        boto_lambda.__init__(self.opts)
        del self.opts
        # Set up MagicMock to replace the boto3 session
        # connections keep getting cached from prior tests, can't find the
        # correct context object to clear it. So randomize the cache key, to prevent any
        # cache hits
        conn_parameters["key"] = "".join(
            random.choice(string.ascii_lowercase + string.digits) for _ in range(50)
        )

        self.patcher = patch("boto3.session.Session")
        self.addCleanup(self.patcher.stop)
        self.addCleanup(delattr, self, "patcher")
        mock_session = self.patcher.start()

        session_instance = mock_session.return_value
        self.conn = MagicMock()
        session_instance.client.return_value = self.conn
        self.addCleanup(delattr, self, "conn")


class TempZipFile:
    def __enter__(self):
        with NamedTemporaryFile(
            suffix=".zip", prefix="salt_test_", delete=False
        ) as tmp:
            to_write = "###\n"
            to_write = salt.utils.stringutils.to_bytes(to_write)
            tmp.write(to_write)
            self.zipfile = tmp.name
        return self.zipfile

    def __exit__(self, type, value, traceback):
        os.remove(self.zipfile)


class BotoLambdaTestCaseMixin:
    pass


class BotoLambdaFunctionTestCase(BotoLambdaTestCaseBase, BotoLambdaTestCaseMixin):
    """
    TestCase for salt.modules.boto_lambda module
    """

    def test_that_when_checking_if_a_function_exists_and_a_function_exists_the_function_exists_method_returns_true(
        self,
    ):
        """
        Tests checking lambda function existence when the lambda function already exists
        """
        self.conn.list_functions.return_value = {"Functions": [function_ret]}
        func_exists_result = boto_lambda.function_exists(
            FunctionName=function_ret["FunctionName"], **conn_parameters
        )

        self.assertTrue(func_exists_result["exists"])

    def test_that_when_checking_if_a_function_exists_and_a_function_does_not_exist_the_function_exists_method_returns_false(
        self,
    ):
        """
        Tests checking lambda function existence when the lambda function does not exist
        """
        self.conn.list_functions.return_value = {"Functions": [function_ret]}
        func_exists_result = boto_lambda.function_exists(
            FunctionName="myfunc", **conn_parameters
        )

        self.assertFalse(func_exists_result["exists"])

    def test_that_when_checking_if_a_function_exists_and_boto3_returns_an_error_the_function_exists_method_returns_error(
        self,
    ):
        """
        Tests checking lambda function existence when boto returns an error
        """
        self.conn.list_functions.side_effect = ClientError(
            error_content, "list_functions"
        )
        func_exists_result = boto_lambda.function_exists(
            FunctionName="myfunc", **conn_parameters
        )

        self.assertEqual(
            func_exists_result.get("error", {}).get("message"),
            error_message.format("list_functions"),
        )

    def test_that_when_creating_a_function_from_zipfile_succeeds_the_create_function_method_returns_true(
        self,
    ):
        """
        tests True function created.
        """
        with patch.dict(
            boto_lambda.__salt__,
            {"boto_iam.get_account_id": MagicMock(return_value="1234")},
        ):
            with TempZipFile() as zipfile:
                self.conn.create_function.return_value = function_ret
                lambda_creation_result = boto_lambda.create_function(
                    FunctionName="testfunction",
                    Runtime="python2.7",
                    Role="myrole",
                    Handler="file.method",
                    ZipFile=zipfile,
                    **conn_parameters
                )

        self.assertTrue(lambda_creation_result["created"])

    def test_that_when_creating_a_function_from_s3_succeeds_the_create_function_method_returns_true(
        self,
    ):
        """
        tests True function created.
        """
        with patch.dict(
            boto_lambda.__salt__,
            {"boto_iam.get_account_id": MagicMock(return_value="1234")},
        ):
            self.conn.create_function.return_value = function_ret
            lambda_creation_result = boto_lambda.create_function(
                FunctionName="testfunction",
                Runtime="python2.7",
                Role="myrole",
                Handler="file.method",
                S3Bucket="bucket",
                S3Key="key",
                **conn_parameters
            )

        self.assertTrue(lambda_creation_result["created"])

    def test_that_when_creating_a_function_without_code_raises_a_salt_invocation_error(
        self,
    ):
        """
        tests Creating a function without code
        """
        with patch.dict(
            boto_lambda.__salt__,
            {"boto_iam.get_account_id": MagicMock(return_value="1234")},
        ):
            with self.assertRaisesRegex(
                SaltInvocationError,
                "Either ZipFile must be specified, or S3Bucket and S3Key must be"
                " provided.",
            ):
                lambda_creation_result = boto_lambda.create_function(
                    FunctionName="testfunction",
                    Runtime="python2.7",
                    Role="myrole",
                    Handler="file.method",
                    **conn_parameters
                )

    def test_that_when_creating_a_function_with_zipfile_and_s3_raises_a_salt_invocation_error(
        self,
    ):
        """
        tests Creating a function without code
        """
        with patch.dict(
            boto_lambda.__salt__,
            {"boto_iam.get_account_id": MagicMock(return_value="1234")},
        ):
            with self.assertRaisesRegex(
                SaltInvocationError,
                "Either ZipFile must be specified, or S3Bucket and S3Key must be"
                " provided.",
            ):
                with TempZipFile() as zipfile:
                    lambda_creation_result = boto_lambda.create_function(
                        FunctionName="testfunction",
                        Runtime="python2.7",
                        Role="myrole",
                        Handler="file.method",
                        ZipFile=zipfile,
                        S3Bucket="bucket",
                        S3Key="key",
                        **conn_parameters
                    )

    def test_that_when_creating_a_function_fails_the_create_function_method_returns_error(
        self,
    ):
        """
        tests False function not created.
        """
        with patch.dict(
            boto_lambda.__salt__,
            {"boto_iam.get_account_id": MagicMock(return_value="1234")},
        ):
            self.conn.create_function.side_effect = ClientError(
                error_content, "create_function"
            )
            with TempZipFile() as zipfile:
                lambda_creation_result = boto_lambda.create_function(
                    FunctionName="testfunction",
                    Runtime="python2.7",
                    Role="myrole",
                    Handler="file.method",
                    ZipFile=zipfile,
                    **conn_parameters
                )
        self.assertEqual(
            lambda_creation_result.get("error", {}).get("message"),
            error_message.format("create_function"),
        )

    def test_that_when_deleting_a_function_succeeds_the_delete_function_method_returns_true(
        self,
    ):
        """
        tests True function deleted.
        """
        with patch.dict(
            boto_lambda.__salt__,
            {"boto_iam.get_account_id": MagicMock(return_value="1234")},
        ):
            result = boto_lambda.delete_function(
                FunctionName="testfunction", Qualifier=1, **conn_parameters
            )

        self.assertTrue(result["deleted"])

    def test_that_when_deleting_a_function_fails_the_delete_function_method_returns_false(
        self,
    ):
        """
        tests False function not deleted.
        """
        with patch.dict(
            boto_lambda.__salt__,
            {"boto_iam.get_account_id": MagicMock(return_value="1234")},
        ):
            self.conn.delete_function.side_effect = ClientError(
                error_content, "delete_function"
            )
            result = boto_lambda.delete_function(
                FunctionName="testfunction", **conn_parameters
            )
        self.assertFalse(result["deleted"])

    def test_that_when_describing_function_it_returns_the_dict_of_properties_returns_true(
        self,
    ):
        """
        Tests describing parameters if function exists
        """
        self.conn.list_functions.return_value = {"Functions": [function_ret]}

        with patch.dict(
            boto_lambda.__salt__,
            {"boto_iam.get_account_id": MagicMock(return_value="1234")},
        ):
            result = boto_lambda.describe_function(
                FunctionName=function_ret["FunctionName"], **conn_parameters
            )

        self.assertEqual(result, {"function": function_ret})

    def test_that_when_describing_function_it_returns_the_dict_of_properties_returns_false(
        self,
    ):
        """
        Tests describing parameters if function does not exist
        """
        self.conn.list_functions.return_value = {"Functions": []}
        with patch.dict(
            boto_lambda.__salt__,
            {"boto_iam.get_account_id": MagicMock(return_value="1234")},
        ):
            result = boto_lambda.describe_function(
                FunctionName="testfunction", **conn_parameters
            )

        self.assertFalse(result["function"])

    def test_that_when_describing_lambda_on_client_error_it_returns_error(self):
        """
        Tests describing parameters failure
        """
        self.conn.list_functions.side_effect = ClientError(
            error_content, "list_functions"
        )
        result = boto_lambda.describe_function(
            FunctionName="testfunction", **conn_parameters
        )
        self.assertTrue("error" in result)

    def test_that_when_updating_a_function_succeeds_the_update_function_method_returns_true(
        self,
    ):
        """
        tests True function updated.
        """
        with patch.dict(
            boto_lambda.__salt__,
            {"boto_iam.get_account_id": MagicMock(return_value="1234")},
        ):
            self.conn.update_function_config.return_value = function_ret
            result = boto_lambda.update_function_config(
                FunctionName=function_ret["FunctionName"],
                Role="myrole",
                **conn_parameters
            )

        self.assertTrue(result["updated"])

    def test_that_when_updating_a_function_fails_the_update_function_method_returns_error(
        self,
    ):
        """
        tests False function not updated.
        """
        with patch.dict(
            boto_lambda.__salt__,
            {"boto_iam.get_account_id": MagicMock(return_value="1234")},
        ):
            self.conn.update_function_configuration.side_effect = ClientError(
                error_content, "update_function"
            )
            result = boto_lambda.update_function_config(
                FunctionName="testfunction", Role="myrole", **conn_parameters
            )
        self.assertEqual(
            result.get("error", {}).get("message"),
            error_message.format("update_function"),
        )

    def test_that_when_updating_function_code_from_zipfile_succeeds_the_update_function_method_returns_true(
        self,
    ):
        """
        tests True function updated.
        """
        with patch.dict(
            boto_lambda.__salt__,
            {"boto_iam.get_account_id": MagicMock(return_value="1234")},
        ):
            with TempZipFile() as zipfile:
                self.conn.update_function_code.return_value = function_ret
                result = boto_lambda.update_function_code(
                    FunctionName=function_ret["FunctionName"],
                    ZipFile=zipfile,
                    **conn_parameters
                )

        self.assertTrue(result["updated"])

    def test_that_when_updating_function_code_from_s3_succeeds_the_update_function_method_returns_true(
        self,
    ):
        """
        tests True function updated.
        """
        with patch.dict(
            boto_lambda.__salt__,
            {"boto_iam.get_account_id": MagicMock(return_value="1234")},
        ):
            self.conn.update_function_code.return_value = function_ret
            result = boto_lambda.update_function_code(
                FunctionName="testfunction",
                S3Bucket="bucket",
                S3Key="key",
                **conn_parameters
            )

        self.assertTrue(result["updated"])

    def test_that_when_updating_function_code_without_code_raises_a_salt_invocation_error(
        self,
    ):
        """
        tests Creating a function without code
        """
        with patch.dict(
            boto_lambda.__salt__,
            {"boto_iam.get_account_id": MagicMock(return_value="1234")},
        ):
            with self.assertRaisesRegex(
                SaltInvocationError,
                "Either ZipFile must be specified, or S3Bucket "
                "and S3Key must be provided.",
            ):
                result = boto_lambda.update_function_code(
                    FunctionName="testfunction", **conn_parameters
                )

    def test_that_when_updating_function_code_fails_the_update_function_method_returns_error(
        self,
    ):
        """
        tests False function not updated.
        """
        with patch.dict(
            boto_lambda.__salt__,
            {"boto_iam.get_account_id": MagicMock(return_value="1234")},
        ):
            self.conn.update_function_code.side_effect = ClientError(
                error_content, "update_function_code"
            )
            result = boto_lambda.update_function_code(
                FunctionName="testfunction",
                S3Bucket="bucket",
                S3Key="key",
                **conn_parameters
            )
        self.assertEqual(
            result.get("error", {}).get("message"),
            error_message.format("update_function_code"),
        )

    def test_that_when_listing_function_versions_succeeds_the_list_function_versions_method_returns_true(
        self,
    ):
        """
        tests True function versions listed.
        """
        with patch.dict(
            boto_lambda.__salt__,
            {"boto_iam.get_account_id": MagicMock(return_value="1234")},
        ):
            self.conn.list_versions_by_function.return_value = {
                "Versions": [function_ret]
            }
            result = boto_lambda.list_function_versions(
                FunctionName="testfunction", **conn_parameters
            )

        self.assertTrue(result["Versions"])

    def test_that_when_listing_function_versions_fails_the_list_function_versions_method_returns_false(
        self,
    ):
        """
        tests False no function versions listed.
        """
        with patch.dict(
            boto_lambda.__salt__,
            {"boto_iam.get_account_id": MagicMock(return_value="1234")},
        ):
            self.conn.list_versions_by_function.return_value = {"Versions": []}
            result = boto_lambda.list_function_versions(
                FunctionName="testfunction", **conn_parameters
            )
        self.assertFalse(result["Versions"])

    def test_that_when_listing_function_versions_fails_the_list_function_versions_method_returns_error(
        self,
    ):
        """
        tests False function versions error.
        """
        with patch.dict(
            boto_lambda.__salt__,
            {"boto_iam.get_account_id": MagicMock(return_value="1234")},
        ):
            self.conn.list_versions_by_function.side_effect = ClientError(
                error_content, "list_versions_by_function"
            )
            result = boto_lambda.list_function_versions(
                FunctionName="testfunction", **conn_parameters
            )
        self.assertEqual(
            result.get("error", {}).get("message"),
            error_message.format("list_versions_by_function"),
        )


@pytest.mark.skipif(HAS_BOTO is False, reason="The boto module must be installed.")
@pytest.mark.skipif(
    _has_required_boto() is False,
    reason="The boto3 module must be greater than or equal to version {}".format(
        required_boto3_version
    ),
)
class BotoLambdaAliasTestCase(BotoLambdaTestCaseBase, BotoLambdaTestCaseMixin):
    """
    TestCase for salt.modules.boto_lambda module aliases
    """

    def test_that_when_creating_an_alias_succeeds_the_create_alias_method_returns_true(
        self,
    ):
        """
        tests True alias created.
        """
        self.conn.create_alias.return_value = alias_ret
        result = boto_lambda.create_alias(
            FunctionName="testfunction",
            Name=alias_ret["Name"],
            FunctionVersion=alias_ret["FunctionVersion"],
            **conn_parameters
        )

        self.assertTrue(result["created"])

    def test_that_when_creating_an_alias_fails_the_create_alias_method_returns_error(
        self,
    ):
        """
        tests False alias not created.
        """
        self.conn.create_alias.side_effect = ClientError(error_content, "create_alias")
        result = boto_lambda.create_alias(
            FunctionName="testfunction",
            Name=alias_ret["Name"],
            FunctionVersion=alias_ret["FunctionVersion"],
            **conn_parameters
        )
        self.assertEqual(
            result.get("error", {}).get("message"), error_message.format("create_alias")
        )

    def test_that_when_deleting_an_alias_succeeds_the_delete_alias_method_returns_true(
        self,
    ):
        """
        tests True alias deleted.
        """
        result = boto_lambda.delete_alias(
            FunctionName="testfunction", Name=alias_ret["Name"], **conn_parameters
        )

        self.assertTrue(result["deleted"])

    def test_that_when_deleting_an_alias_fails_the_delete_alias_method_returns_false(
        self,
    ):
        """
        tests False alias not deleted.
        """
        self.conn.delete_alias.side_effect = ClientError(error_content, "delete_alias")
        result = boto_lambda.delete_alias(
            FunctionName="testfunction", Name=alias_ret["Name"], **conn_parameters
        )
        self.assertFalse(result["deleted"])

    def test_that_when_checking_if_an_alias_exists_and_the_alias_exists_the_alias_exists_method_returns_true(
        self,
    ):
        """
        Tests checking lambda alias existence when the lambda alias already exists
        """
        self.conn.list_aliases.return_value = {"Aliases": [alias_ret]}
        result = boto_lambda.alias_exists(
            FunctionName="testfunction", Name=alias_ret["Name"], **conn_parameters
        )
        self.assertTrue(result["exists"])

    def test_that_when_checking_if_an_alias_exists_and_the_alias_does_not_exist_the_alias_exists_method_returns_false(
        self,
    ):
        """
        Tests checking lambda alias existence when the lambda alias does not exist
        """
        self.conn.list_aliases.return_value = {"Aliases": [alias_ret]}
        result = boto_lambda.alias_exists(
            FunctionName="testfunction", Name="otheralias", **conn_parameters
        )

        self.assertFalse(result["exists"])

    def test_that_when_checking_if_an_alias_exists_and_boto3_returns_an_error_the_alias_exists_method_returns_error(
        self,
    ):
        """
        Tests checking lambda alias existence when boto returns an error
        """
        self.conn.list_aliases.side_effect = ClientError(error_content, "list_aliases")
        result = boto_lambda.alias_exists(
            FunctionName="testfunction", Name=alias_ret["Name"], **conn_parameters
        )

        self.assertEqual(
            result.get("error", {}).get("message"), error_message.format("list_aliases")
        )

    def test_that_when_describing_alias_it_returns_the_dict_of_properties_returns_true(
        self,
    ):
        """
        Tests describing parameters if alias exists
        """
        self.conn.list_aliases.return_value = {"Aliases": [alias_ret]}

        result = boto_lambda.describe_alias(
            FunctionName="testfunction", Name=alias_ret["Name"], **conn_parameters
        )

        self.assertEqual(result, {"alias": alias_ret})

    def test_that_when_describing_alias_it_returns_the_dict_of_properties_returns_false(
        self,
    ):
        """
        Tests describing parameters if alias does not exist
        """
        self.conn.list_aliases.return_value = {"Aliases": [alias_ret]}
        result = boto_lambda.describe_alias(
            FunctionName="testfunction", Name="othername", **conn_parameters
        )

        self.assertFalse(result["alias"])

    def test_that_when_describing_lambda_on_client_error_it_returns_error(self):
        """
        Tests describing parameters failure
        """
        self.conn.list_aliases.side_effect = ClientError(error_content, "list_aliases")
        result = boto_lambda.describe_alias(
            FunctionName="testfunction", Name=alias_ret["Name"], **conn_parameters
        )
        self.assertTrue("error" in result)

    def test_that_when_updating_an_alias_succeeds_the_update_alias_method_returns_true(
        self,
    ):
        """
        tests True alias updated.
        """
        self.conn.update_alias.return_value = alias_ret
        result = boto_lambda.update_alias(
            FunctionName="testfunctoin",
            Name=alias_ret["Name"],
            Description=alias_ret["Description"],
            **conn_parameters
        )

        self.assertTrue(result["updated"])

    def test_that_when_updating_an_alias_fails_the_update_alias_method_returns_error(
        self,
    ):
        """
        tests False alias not updated.
        """
        self.conn.update_alias.side_effect = ClientError(error_content, "update_alias")
        result = boto_lambda.update_alias(
            FunctionName="testfunction", Name=alias_ret["Name"], **conn_parameters
        )
        self.assertEqual(
            result.get("error", {}).get("message"), error_message.format("update_alias")
        )


@pytest.mark.skipif(HAS_BOTO is False, reason="The boto module must be installed.")
@pytest.mark.skipif(
    _has_required_boto() is False,
    reason="The boto3 module must be greater than or equal to version {}".format(
        required_boto3_version
    ),
)
class BotoLambdaEventSourceMappingTestCase(
    BotoLambdaTestCaseBase, BotoLambdaTestCaseMixin
):
    """
    TestCase for salt.modules.boto_lambda module mappings
    """

    def test_that_when_creating_a_mapping_succeeds_the_create_event_source_mapping_method_returns_true(
        self,
    ):
        """
        tests True mapping created.
        """
        self.conn.create_event_source_mapping.return_value = event_source_mapping_ret
        result = boto_lambda.create_event_source_mapping(
            EventSourceArn=event_source_mapping_ret["EventSourceArn"],
            FunctionName=event_source_mapping_ret["FunctionArn"],
            StartingPosition="LATEST",
            **conn_parameters
        )

        self.assertTrue(result["created"])

    def test_that_when_creating_an_event_source_mapping_fails_the_create_event_source_mapping_method_returns_error(
        self,
    ):
        """
        tests False mapping not created.
        """
        self.conn.create_event_source_mapping.side_effect = ClientError(
            error_content, "create_event_source_mapping"
        )
        result = boto_lambda.create_event_source_mapping(
            EventSourceArn=event_source_mapping_ret["EventSourceArn"],
            FunctionName=event_source_mapping_ret["FunctionArn"],
            StartingPosition="LATEST",
            **conn_parameters
        )
        self.assertEqual(
            result.get("error", {}).get("message"),
            error_message.format("create_event_source_mapping"),
        )

    def test_that_when_listing_mapping_ids_succeeds_the_get_event_source_mapping_ids_method_returns_true(
        self,
    ):
        """
        tests True mapping ids listed.
        """
        self.conn.list_event_source_mappings.return_value = {
            "EventSourceMappings": [event_source_mapping_ret]
        }
        result = boto_lambda.get_event_source_mapping_ids(
            EventSourceArn=event_source_mapping_ret["EventSourceArn"],
            FunctionName=event_source_mapping_ret["FunctionArn"],
            **conn_parameters
        )

        self.assertTrue(result)

    def test_that_when_listing_event_source_mapping_ids_fails_the_get_event_source_mapping_ids_versions_method_returns_false(
        self,
    ):
        """
        tests False no mapping ids listed.
        """
        self.conn.list_event_source_mappings.return_value = {"EventSourceMappings": []}
        result = boto_lambda.get_event_source_mapping_ids(
            EventSourceArn=event_source_mapping_ret["EventSourceArn"],
            FunctionName=event_source_mapping_ret["FunctionArn"],
            **conn_parameters
        )
        self.assertFalse(result)

    def test_that_when_listing_event_source_mapping_ids_fails_the_get_event_source_mapping_ids_method_returns_error(
        self,
    ):
        """
        tests False mapping ids error.
        """
        self.conn.list_event_source_mappings.side_effect = ClientError(
            error_content, "list_event_source_mappings"
        )
        result = boto_lambda.get_event_source_mapping_ids(
            EventSourceArn=event_source_mapping_ret["EventSourceArn"],
            FunctionName=event_source_mapping_ret["FunctionArn"],
            **conn_parameters
        )
        self.assertEqual(
            result.get("error", {}).get("message"),
            error_message.format("list_event_source_mappings"),
        )

    def test_that_when_deleting_an_event_source_mapping_by_UUID_succeeds_the_delete_event_source_mapping_method_returns_true(
        self,
    ):
        """
        tests True mapping deleted.
        """
        result = boto_lambda.delete_event_source_mapping(
            UUID=event_source_mapping_ret["UUID"], **conn_parameters
        )
        self.assertTrue(result["deleted"])

    @pytest.mark.skip(
        reason="This appears to leak memory and crash the unit test suite"
    )
    def test_that_when_deleting_an_event_source_mapping_by_name_succeeds_the_delete_event_source_mapping_method_returns_true(
        self,
    ):
        """
        tests True mapping deleted.
        """
        self.conn.list_event_source_mappings.return_value = {
            "EventSourceMappings": [event_source_mapping_ret]
        }
        result = boto_lambda.delete_event_source_mapping(
            EventSourceArn=event_source_mapping_ret["EventSourceArn"],
            FunctionName=event_source_mapping_ret["FunctionArn"],
            **conn_parameters
        )
        self.assertTrue(result["deleted"])

    def test_that_when_deleting_an_event_source_mapping_without_identifier_the_delete_event_source_mapping_method_raises_saltinvocationexception(
        self,
    ):
        """
        tests Deleting a mapping without identifier
        """
        with self.assertRaisesRegex(
            SaltInvocationError,
            "Either UUID must be specified, or EventSourceArn "
            "and FunctionName must be provided.",
        ):
            result = boto_lambda.delete_event_source_mapping(**conn_parameters)

    def test_that_when_deleting_an_event_source_mapping_fails_the_delete_event_source_mapping_method_returns_false(
        self,
    ):
        """
        tests False mapping not deleted.
        """
        self.conn.delete_event_source_mapping.side_effect = ClientError(
            error_content, "delete_event_source_mapping"
        )
        result = boto_lambda.delete_event_source_mapping(
            UUID=event_source_mapping_ret["UUID"], **conn_parameters
        )
        self.assertFalse(result["deleted"])

    def test_that_when_checking_if_an_event_source_mapping_exists_and_the_event_source_mapping_exists_the_event_source_mapping_exists_method_returns_true(
        self,
    ):
        """
        Tests checking lambda event_source_mapping existence when the lambda
        event_source_mapping already exists
        """
        self.conn.get_event_source_mapping.return_value = event_source_mapping_ret
        result = boto_lambda.event_source_mapping_exists(
            UUID=event_source_mapping_ret["UUID"], **conn_parameters
        )
        self.assertTrue(result["exists"])

    def test_that_when_checking_if_an_event_source_mapping_exists_and_the_event_source_mapping_does_not_exist_the_event_source_mapping_exists_method_returns_false(
        self,
    ):
        """
        Tests checking lambda event_source_mapping existence when the lambda
        event_source_mapping does not exist
        """
        self.conn.get_event_source_mapping.return_value = None
        result = boto_lambda.event_source_mapping_exists(
            UUID="other_UUID", **conn_parameters
        )
        self.assertFalse(result["exists"])

    def test_that_when_checking_if_an_event_source_mapping_exists_and_boto3_returns_an_error_the_event_source_mapping_exists_method_returns_error(
        self,
    ):
        """
        Tests checking lambda event_source_mapping existence when boto returns an error
        """
        self.conn.get_event_source_mapping.side_effect = ClientError(
            error_content, "list_event_source_mappings"
        )
        result = boto_lambda.event_source_mapping_exists(
            UUID=event_source_mapping_ret["UUID"], **conn_parameters
        )
        self.assertEqual(
            result.get("error", {}).get("message"),
            error_message.format("list_event_source_mappings"),
        )

    def test_that_when_describing_event_source_mapping_it_returns_the_dict_of_properties_returns_true(
        self,
    ):
        """
        Tests describing parameters if event_source_mapping exists
        """
        self.conn.get_event_source_mapping.return_value = event_source_mapping_ret
        result = boto_lambda.describe_event_source_mapping(
            UUID=event_source_mapping_ret["UUID"], **conn_parameters
        )
        self.assertEqual(result, {"event_source_mapping": event_source_mapping_ret})

    def test_that_when_describing_event_source_mapping_it_returns_the_dict_of_properties_returns_false(
        self,
    ):
        """
        Tests describing parameters if event_source_mapping does not exist
        """
        self.conn.get_event_source_mapping.return_value = None
        result = boto_lambda.describe_event_source_mapping(
            UUID=event_source_mapping_ret["UUID"], **conn_parameters
        )
        self.assertFalse(result["event_source_mapping"])

    def test_that_when_describing_event_source_mapping_on_client_error_it_returns_error(
        self,
    ):
        """
        Tests describing parameters failure
        """
        self.conn.get_event_source_mapping.side_effect = ClientError(
            error_content, "get_event_source_mapping"
        )
        result = boto_lambda.describe_event_source_mapping(
            UUID=event_source_mapping_ret["UUID"], **conn_parameters
        )
        self.assertTrue("error" in result)

    def test_that_when_updating_an_event_source_mapping_succeeds_the_update_event_source_mapping_method_returns_true(
        self,
    ):
        """
        tests True event_source_mapping updated.
        """
        self.conn.update_event_source_mapping.return_value = event_source_mapping_ret
        result = boto_lambda.update_event_source_mapping(
            UUID=event_source_mapping_ret["UUID"],
            FunctionName=event_source_mapping_ret["FunctionArn"],
            **conn_parameters
        )

        self.assertTrue(result["updated"])

    def test_that_when_updating_an_event_source_mapping_fails_the_update_event_source_mapping_method_returns_error(
        self,
    ):
        """
        tests False event_source_mapping not updated.
        """
        self.conn.update_event_source_mapping.side_effect = ClientError(
            error_content, "update_event_source_mapping"
        )
        result = boto_lambda.update_event_source_mapping(
            UUID=event_source_mapping_ret["UUID"],
            FunctionName=event_source_mapping_ret["FunctionArn"],
            **conn_parameters
        )
        self.assertEqual(
            result.get("error", {}).get("message"),
            error_message.format("update_event_source_mapping"),
        )



"""
Fileserver backend which serves files pushed to the Master

The :mod:`cp.push <salt.modules.cp.push>` function allows Minions to push files
up to the Master. Using this backend, these pushed files are exposed to other
Minions via the Salt fileserver.

To enable minionfs, :conf_master:`file_recv` needs to be set to ``True`` in the
master config file (otherwise :mod:`cp.push <salt.modules.cp.push>` will not be
allowed to push files to the Master), and ``minionfs`` must be added to the
:conf_master:`fileserver_backends` list.

.. code-block:: yaml

    fileserver_backend:
      - minionfs

.. note::
    ``minion`` also works here. Prior to the 2018.3.0 release, *only*
    ``minion`` would work.

Other minionfs settings include: :conf_master:`minionfs_whitelist`,
:conf_master:`minionfs_blacklist`, :conf_master:`minionfs_mountpoint`, and
:conf_master:`minionfs_env`.

.. seealso:: :ref:`tutorial-minionfs`

"""

import logging
import os

import salt.fileserver
import salt.utils.files
import salt.utils.gzip_util
import salt.utils.hashutils
import salt.utils.path
import salt.utils.stringutils
import salt.utils.url
import salt.utils.versions

log = logging.getLogger(__name__)


# Define the module's virtual name
__virtualname__ = "minionfs"


def __virtual__():
    """
    Only load if file_recv is enabled
    """
    if __virtualname__ not in __opts__["fileserver_backend"]:
        return False
    return __virtualname__ if __opts__["file_recv"] else False


def _is_exposed(minion):
    """
    Check if the minion is exposed, based on the whitelist and blacklist
    """
    return salt.utils.stringutils.check_whitelist_blacklist(
        minion,
        whitelist=__opts__["minionfs_whitelist"],
        blacklist=__opts__["minionfs_blacklist"],
    )


def find_file(path, tgt_env="base", **kwargs):  # pylint: disable=W0613
    """
    Search the environment for the relative path
    """
    fnd = {"path": "", "rel": ""}
    if os.path.isabs(path):
        return fnd
    if tgt_env not in envs():
        return fnd
    if os.path.basename(path) == "top.sls":
        log.debug(
            "minionfs will NOT serve top.sls for security reasons (path requested: %s)",
            path,
        )
        return fnd

    mountpoint = salt.utils.url.strip_proto(__opts__["minionfs_mountpoint"])
    # Remove the mountpoint to get the "true" path
    path = path[len(mountpoint) :].lstrip(os.path.sep)
    try:
        minion, pushed_file = path.split(os.sep, 1)
    except ValueError:
        return fnd
    if not _is_exposed(minion):
        return fnd
    full = os.path.join(__opts__["cachedir"], "minions", minion, "files", pushed_file)
    if os.path.isfile(full) and not salt.fileserver.is_file_ignored(__opts__, full):
        fnd["path"] = full
        fnd["rel"] = path
        fnd["stat"] = list(os.stat(full))
        return fnd
    return fnd


def envs():
    """
    Returns the one environment specified for minionfs in the master
    configuration.
    """
    return [__opts__["minionfs_env"]]


def serve_file(load, fnd):
    """
    Return a chunk from a file based on the data received

    CLI Example:

    .. code-block:: bash

        # Push the file to the master
        $ salt 'source-minion' cp.push /path/to/the/file
        $ salt 'destination-minion' cp.get_file salt://source-minion/path/to/the/file /destination/file
    """
    ret = {"data": "", "dest": ""}
    if not fnd["path"]:
        return ret
    ret["dest"] = fnd["rel"]
    gzip = load.get("gzip", None)
    fpath = os.path.normpath(fnd["path"])

    # AP
    # May I sleep here to slow down serving of big files?
    # How many threads are serving files?
    with salt.utils.files.fopen(fpath, "rb") as fp_:
        fp_.seek(load["loc"])
        data = fp_.read(__opts__["file_buffer_size"])
        if data and not salt.utils.files.is_binary(fpath):
            data = data.decode(__salt_system_encoding__)
        if gzip and data:
            data = salt.utils.gzip_util.compress(data, gzip)
            ret["gzip"] = gzip
        ret["data"] = data
    return ret


def update():
    """
    When we are asked to update (regular interval) lets reap the cache
    """
    try:
        salt.fileserver.reap_fileserver_cache_dir(
            os.path.join(__opts__["cachedir"], "minionfs/hash"), find_file
        )
    except os.error:
        # Hash file won't exist if no files have yet been served up
        pass


def file_hash(load, fnd):
    """
    Return a file hash, the hash type is set in the master config file
    """
    path = fnd["path"]
    ret = {}

    if "env" in load:
        # "env" is not supported; Use "saltenv".
        load.pop("env")

    if load["saltenv"] not in envs():
        return {}

    # if the file doesn't exist, we can't get a hash
    if not path or not os.path.isfile(path):
        return ret

    # set the hash_type as it is determined by config-- so mechanism won't change that
    ret["hash_type"] = __opts__["hash_type"]

    # check if the hash is cached
    # cache file's contents should be "hash:mtime"
    cache_path = os.path.join(
        __opts__["cachedir"],
        "minionfs",
        "hash",
        load["saltenv"],
        "{}.hash.{}".format(fnd["rel"], __opts__["hash_type"]),
    )
    # if we have a cache, serve that if the mtime hasn't changed
    if os.path.exists(cache_path):
        try:
            with salt.utils.files.fopen(cache_path, "rb") as fp_:
                try:
                    hsum, mtime = salt.utils.stringutils.to_unicode(fp_.read()).split(
                        ":"
                    )
                except ValueError:
                    log.debug(
                        "Fileserver attempted to read incomplete cache file. Retrying."
                    )
                    file_hash(load, fnd)
                    return ret
                if os.path.getmtime(path) == mtime:
                    # check if mtime changed
                    ret["hsum"] = hsum
                    return ret
        # Can't use Python select() because we need Windows support
        except os.error:
            log.debug("Fileserver encountered lock when reading cache file. Retrying.")
            file_hash(load, fnd)
            return ret

    # if we don't have a cache entry-- lets make one
    ret["hsum"] = salt.utils.hashutils.get_hash(path, __opts__["hash_type"])
    cache_dir = os.path.dirname(cache_path)
    # make cache directory if it doesn't exist
    if not os.path.exists(cache_dir):
        os.makedirs(cache_dir)
    # save the cache object "hash:mtime"
    cache_object = "{}:{}".format(ret["hsum"], os.path.getmtime(path))
    with salt.utils.files.flopen(cache_path, "w") as fp_:
        fp_.write(cache_object)
    return ret


def file_list(load):
    """
    Return a list of all files on the file server in a specified environment
    """
    if "env" in load:
        # "env" is not supported; Use "saltenv".
        load.pop("env")

    if load["saltenv"] not in envs():
        return []
    mountpoint = salt.utils.url.strip_proto(__opts__["minionfs_mountpoint"])
    prefix = load.get("prefix", "").strip("/")
    if mountpoint and prefix.startswith(mountpoint + os.path.sep):
        prefix = prefix[len(mountpoint + os.path.sep) :]

    minions_cache_dir = os.path.join(__opts__["cachedir"], "minions")
    minion_dirs = os.listdir(minions_cache_dir)

    # If the prefix is not an empty string, then get the minion id from it. The
    # minion ID will be the part before the first slash, so if there is no
    # slash, this is an invalid path.
    if prefix:
        tgt_minion, _, prefix = prefix.partition("/")
        if not prefix:
            # No minion ID in path
            return []
        # Reassign minion_dirs so we don't unnecessarily walk every minion's
        # pushed files
        if tgt_minion not in minion_dirs:
            log.warning(
                "No files found in minionfs cache for minion ID '%s'", tgt_minion
            )
            return []
        minion_dirs = [tgt_minion]

    ret = []
    for minion in minion_dirs:
        if not _is_exposed(minion):
            continue
        minion_files_dir = os.path.join(minions_cache_dir, minion, "files")
        if not os.path.isdir(minion_files_dir):
            log.debug(
                "minionfs: could not find files directory under %s!",
                os.path.join(minions_cache_dir, minion),
            )
            continue
        walk_dir = os.path.join(minion_files_dir, prefix)
        # Do not follow links for security reasons
        for root, _, files in salt.utils.path.os_walk(walk_dir, followlinks=False):
            for fname in files:
                # Ignore links for security reasons
                if os.path.islink(os.path.join(root, fname)):
                    continue
                relpath = os.path.relpath(os.path.join(root, fname), minion_files_dir)
                if relpath.startswith("../"):
                    continue
                rel_fn = os.path.join(mountpoint, minion, relpath)
                if not salt.fileserver.is_file_ignored(__opts__, rel_fn):
                    ret.append(rel_fn)
    return ret


# There should be no emptydirs
# def file_list_emptydirs(load):


def dir_list(load):
    """
    Return a list of all directories on the master

    CLI Example:

    .. code-block:: bash

        $ salt 'source-minion' cp.push /absolute/path/file  # Push the file to the master
        $ salt 'destination-minion' cp.list_master_dirs
        destination-minion:
            - source-minion/absolute
            - source-minion/absolute/path
    """
    if "env" in load:
        # "env" is not supported; Use "saltenv".
        load.pop("env")

    if load["saltenv"] not in envs():
        return []
    mountpoint = salt.utils.url.strip_proto(__opts__["minionfs_mountpoint"])
    prefix = load.get("prefix", "").strip("/")
    if mountpoint and prefix.startswith(mountpoint + os.path.sep):
        prefix = prefix[len(mountpoint + os.path.sep) :]

    minions_cache_dir = os.path.join(__opts__["cachedir"], "minions")
    minion_dirs = os.listdir(minions_cache_dir)

    # If the prefix is not an empty string, then get the minion id from it. The
    # minion ID will be the part before the first slash, so if there is no
    # slash, this is an invalid path.
    if prefix:
        tgt_minion, _, prefix = prefix.partition("/")
        if not prefix:
            # No minion ID in path
            return []
        # Reassign minion_dirs so we don't unnecessarily walk every minion's
        # pushed files
        if tgt_minion not in minion_dirs:
            log.warning(
                "No files found in minionfs cache for minion ID '%s'", tgt_minion
            )
            return []
        minion_dirs = [tgt_minion]

    ret = []
    for minion in os.listdir(minions_cache_dir):
        if not _is_exposed(minion):
            continue
        minion_files_dir = os.path.join(minions_cache_dir, minion, "files")
        if not os.path.isdir(minion_files_dir):
            log.warning(
                "minionfs: could not find files directory under %s!",
                os.path.join(minions_cache_dir, minion),
            )
            continue
        walk_dir = os.path.join(minion_files_dir, prefix)
        # Do not follow links for security reasons
        for root, _, _ in salt.utils.path.os_walk(walk_dir, followlinks=False):
            relpath = os.path.relpath(root, minion_files_dir)
            # Ensure that the current directory and directories outside of
            # the minion dir do not end up in return list
            if relpath in (".", "..") or relpath.startswith("../"):
                continue
            ret.append(os.path.join(mountpoint, minion, relpath))
    return ret



"""
    :codeauthor: Denys Havrysh <denys.gavrysh@gmail.com>
"""

import logging
import os
import pprint
import shutil
import tempfile

import salt._logging
import salt.config
import salt.syspaths
import salt.utils.jid
import salt.utils.parsers
import salt.utils.platform
from tests.support.helpers import TstSuiteLoggingHandler
from tests.support.mock import ANY, MagicMock, patch
from tests.support.runtests import RUNTIME_VARS
from tests.support.unit import TestCase

log = logging.getLogger(__name__)


class ErrorMock:  # pylint: disable=too-few-public-methods
    """
    Error handling
    """

    def __init__(self):
        """
        init
        """
        self.msg = None

    def error(self, msg):
        """
        Capture error message
        """
        self.msg = msg


class LogImplMock:
    """
    Logger setup
    """

    def __init__(self):
        """
        init
        """
        self.log_level_console = None
        self.log_file = None
        self.log_level_logfile = None
        self.config = self.original_config = None
        logging_options = salt._logging.get_logging_options_dict()
        if logging_options:
            self.config = logging_options.copy()
            self.original_config = self.config.copy()
        self.temp_log_level = None
        self._console_handler_configured = False
        self._extended_logging_configured = False
        self._logfile_handler_configured = False
        self._real_set_logging_options_dict = salt._logging.set_logging_options_dict
        self._real_get_logging_options_dict = salt._logging.get_logging_options_dict
        self._real_setup_logfile_handler = salt._logging.setup_logfile_handler

    def _destroy(self):
        salt._logging.set_logging_options_dict.__options_dict__ = self.original_config
        salt._logging.shutdown_logfile_handler()

    def setup_temp_handler(self, log_level=None):
        """
        Set temp handler loglevel
        """
        log.debug("Setting temp handler log level to: %s", log_level)
        self.temp_log_level = log_level

    def is_console_handler_configured(self):
        log.debug("Calling is_console_handler_configured")
        return self._console_handler_configured

    def setup_console_handler(
        self, log_level="error", **kwargs
    ):  # pylint: disable=unused-argument
        """
        Set console loglevel
        """
        log.debug("Setting console handler log level to: %s", log_level)
        self.log_level_console = log_level
        self._console_handler_configured = True

    def shutdown_console_handler(self):
        log.debug("Calling shutdown_console_handler")
        self._console_handler_configured = False

    def is_extended_logging_configured(self):
        log.debug("Calling is_extended_logging_configured")
        return self._extended_logging_configured

    def setup_extended_logging(self, opts):
        """
        Set opts
        """
        log.debug("Calling setup_extended_logging")
        self._extended_logging_configured = True

    def shutdown_extended_logging(self):
        log.debug("Calling shutdown_extended_logging")
        self._extended_logging_configured = False

    def is_logfile_handler_configured(self):
        log.debug("Calling is_logfile_handler_configured")
        return self._logfile_handler_configured

    def setup_logfile_handler(
        self, log_path, log_level=None, **kwargs
    ):  # pylint: disable=unused-argument
        """
        Set logfile and loglevel
        """
        log.debug("Setting log file handler path to: %s", log_path)
        log.debug("Setting log file handler log level to: %s", log_level)
        self.log_file = log_path
        self.log_level_logfile = log_level
        self._real_setup_logfile_handler(log_path, log_level=log_level, **kwargs)
        self._logfile_handler_configured = True

    def shutdown_logfile_handler(self):
        log.debug("Calling shutdown_logfile_handler")
        self._logfile_handler_configured = False

    def get_logging_options_dict(self):
        log.debug("Calling get_logging_options_dict")
        return self.config

    def set_logging_options_dict(self, opts):
        log.debug("Calling set_logging_options_dict")
        self._real_set_logging_options_dict(opts)
        self.config = self._real_get_logging_options_dict()
        log.debug("Logging options dict:\n%s", pprint.pformat(self.config))

    def setup_log_granular_levels(self, opts):
        log.debug("Calling setup_log_granular_levels")

    def setup_logging(self):
        log.debug("Mocked setup_logging called")
        # Wether daemonizing or not, either on the main process or on a separate process
        # The log file is going to be configured.
        # The console is the only handler not configured if daemonizing

        # These routines are what happens on salt._logging.setup_logging
        opts = self.get_logging_options_dict()

        if (
            opts.get("configure_console_logger", True)
            and not self.is_console_handler_configured()
        ):
            self.setup_console_handler(
                log_level=opts["log_level"],
                log_format=opts["log_fmt_console"],
                date_format=opts["log_datefmt"],
            )
        if (
            opts.get("configure_file_logger", True)
            and not self.is_logfile_handler_configured()
        ):
            log_file_level = opts["log_level_logfile"] or opts["log_level"]
            if log_file_level != "quiet":
                self.setup_logfile_handler(
                    log_path=opts[opts["log_file_key"]],
                    log_level=log_file_level,
                    log_format=opts["log_fmt_logfile"],
                    date_format=opts["log_datefmt_logfile"],
                    max_bytes=opts["log_rotate_max_bytes"],
                    backup_count=opts["log_rotate_backup_count"],
                    user=opts["user"],
                )
        if not self.is_extended_logging_configured():
            self.setup_extended_logging(opts)
        self.setup_log_granular_levels(opts["log_granular_levels"])


class ObjectView:  # pylint: disable=too-few-public-methods
    """
    Dict object view
    """

    def __init__(self, d):
        self.__dict__ = d


class ParserBase:
    """
    Unit Tests for Log Level Mixin with Salt parsers
    """

    args = []

    log_impl = None

    # Set config option names
    loglevel_config_setting_name = "log_level"
    logfile_config_setting_name = "log_file"
    logfile_loglevel_config_setting_name = (
        "log_level_logfile"  # pylint: disable=invalid-name
    )

    @classmethod
    def setUpClass(cls):
        cls.root_dir = tempfile.mkdtemp(dir=RUNTIME_VARS.TMP)

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(cls.root_dir, ignore_errors=True)

    def setup_log(self):
        """
        Mock logger functions
        """
        testing_config = self.default_config.copy()
        testing_config["root_dir"] = self.root_dir
        for name in ("pki_dir", "cachedir"):
            testing_config[name] = name
        testing_config[self.logfile_config_setting_name] = getattr(
            self, self.logfile_config_setting_name, self.log_file
        )
        self.testing_config = testing_config
        self.addCleanup(setattr, self, "testing_config", None)

        self.log_impl = LogImplMock()
        self.addCleanup(self.log_impl._destroy)
        self.addCleanup(setattr, self, "log_impl", None)

        mocked_functions = {}
        for name in dir(self.log_impl):
            if name.startswith("_"):
                continue
            func = getattr(self.log_impl, name)
            if not callable(func):
                continue
            mocked_functions[name] = func
        patcher = patch.multiple(salt._logging, **mocked_functions)
        patcher.start()
        self.addCleanup(patcher.stop)

    # log level configuration tests

    def test_get_log_level_cli(self):
        """
        Tests that log level match command-line specified value
        """
        # Set defaults
        default_log_level = self.testing_config[self.loglevel_config_setting_name]

        # Set log level in CLI
        log_level = "critical"
        args = ["--log-level", log_level] + self.args

        parser = self.parser()
        with patch(self.config_func, MagicMock(return_value=self.testing_config)):
            parser.parse_args(args)

        console_log_level = getattr(parser.options, self.loglevel_config_setting_name)

        # Check console log level setting
        self.assertEqual(console_log_level, log_level)
        # Check console logger log level
        self.assertEqual(self.log_impl.log_level_console, log_level)
        self.assertEqual(
            self.log_impl.config[self.loglevel_config_setting_name], log_level
        )
        self.assertEqual(self.log_impl.temp_log_level, log_level)
        # Check log file logger log level
        self.assertEqual(self.log_impl.log_level_logfile, default_log_level)

    def test_get_log_level_config(self):
        """
        Tests that log level match the configured value
        """
        args = self.args

        # Set log level in config
        log_level = "info"
        opts = self.testing_config.copy()
        opts.update({self.loglevel_config_setting_name: log_level})

        parser = self.parser()
        with patch(self.config_func, MagicMock(return_value=opts)):
            parser.parse_args(args)

        console_log_level = getattr(parser.options, self.loglevel_config_setting_name)

        # Check console log level setting
        self.assertEqual(console_log_level, log_level)
        # Check console logger log level
        self.assertEqual(self.log_impl.log_level_console, log_level)
        self.assertEqual(
            self.log_impl.config[self.loglevel_config_setting_name], log_level
        )
        self.assertEqual(self.log_impl.temp_log_level, "error")
        # Check log file logger log level
        self.assertEqual(self.log_impl.log_level_logfile, log_level)

    def test_get_log_level_default(self):
        """
        Tests that log level match the default value
        """
        # Set defaults
        log_level = default_log_level = self.testing_config[
            self.loglevel_config_setting_name
        ]

        args = self.args

        parser = self.parser()
        with patch(self.config_func, MagicMock(return_value=self.testing_config)):
            parser.parse_args(args)

        console_log_level = getattr(parser.options, self.loglevel_config_setting_name)

        # Check log level setting
        self.assertEqual(console_log_level, log_level)
        # Check console logger log level
        self.assertEqual(self.log_impl.log_level_console, log_level)
        # Check extended logger
        self.assertEqual(
            self.log_impl.config[self.loglevel_config_setting_name], log_level
        )
        self.assertEqual(self.log_impl.temp_log_level, "error")
        # Check log file logger
        self.assertEqual(self.log_impl.log_level_logfile, default_log_level)
        # Check help message
        self.assertIn(
            "Default: '{}'.".format(default_log_level),
            parser.get_option("--log-level").help,
        )

    # log file configuration tests

    def test_get_log_file_cli(self):
        """
        Tests that log file match command-line specified value
        """
        # Set defaults
        log_level = self.testing_config[self.loglevel_config_setting_name]

        # Set log file in CLI
        log_file = "{}_cli.log".format(self.log_file)
        args = ["--log-file", log_file] + self.args

        parser = self.parser()
        with patch(self.config_func, MagicMock(return_value=self.testing_config)):
            parser.parse_args(args)

        log_file_option = getattr(parser.options, self.logfile_config_setting_name)

        # Check console logger
        self.assertEqual(self.log_impl.log_level_console, log_level)
        # Check extended logger
        self.assertEqual(
            self.log_impl.config[self.loglevel_config_setting_name], log_level
        )
        self.assertEqual(
            self.log_impl.config[self.logfile_config_setting_name], log_file
        )
        # Check temp logger
        self.assertEqual(self.log_impl.temp_log_level, "error")
        # Check log file setting
        self.assertEqual(log_file_option, log_file)
        # Check log file logger
        self.assertEqual(self.log_impl.log_file, log_file)

    def test_get_log_file_config(self):
        """
        Tests that log file match the configured value
        """
        # Set defaults
        log_level = self.testing_config[self.loglevel_config_setting_name]

        args = self.args

        # Set log file in config
        log_file = "{}_config.log".format(self.log_file)
        opts = self.testing_config.copy()
        opts.update({self.logfile_config_setting_name: log_file})

        parser = self.parser()
        with patch(self.config_func, MagicMock(return_value=opts)):
            parser.parse_args(args)

        log_file_option = getattr(parser.options, self.logfile_config_setting_name)

        # Check console logger
        self.assertEqual(self.log_impl.log_level_console, log_level)
        # Check extended logger
        self.assertEqual(
            self.log_impl.config[self.loglevel_config_setting_name], log_level
        )
        self.assertEqual(
            self.log_impl.config[self.logfile_config_setting_name], log_file
        )
        # Check temp logger
        self.assertEqual(self.log_impl.temp_log_level, "error")
        # Check log file setting
        self.assertEqual(log_file_option, log_file)
        # Check log file logger
        self.assertEqual(self.log_impl.log_file, log_file)

    def test_get_log_file_default(self):
        """
        Tests that log file match the default value
        """
        # Set defaults
        log_level = self.testing_config[self.loglevel_config_setting_name]
        log_file = self.testing_config[self.logfile_config_setting_name]
        default_log_file = self.default_config[self.logfile_config_setting_name]

        args = self.args

        parser = self.parser()
        with patch(self.config_func, MagicMock(return_value=self.testing_config)):
            parser.parse_args(args)

        log_file_option = getattr(parser.options, self.logfile_config_setting_name)

        # Check console logger
        self.assertEqual(self.log_impl.log_level_console, log_level)
        # Check extended logger
        self.assertEqual(
            self.log_impl.config[self.loglevel_config_setting_name], log_level
        )
        self.assertEqual(
            self.log_impl.config[self.logfile_config_setting_name], log_file
        )
        # Check temp logger
        self.assertEqual(self.log_impl.temp_log_level, "error")
        # Check log file setting
        self.assertEqual(log_file_option, log_file)
        # Check log file logger
        self.assertEqual(self.log_impl.log_file, log_file)
        # Check help message
        self.assertIn(
            "Default: '{}'.".format(default_log_file),
            parser.get_option("--log-file").help,
        )

    # log file log level configuration tests

    def test_get_log_file_level_cli(self):
        """
        Tests that file log level match command-line specified value
        """
        # Set defaults
        default_log_level = self.testing_config[self.loglevel_config_setting_name]

        # Set log file level in CLI
        log_level_logfile = "error"
        args = ["--log-file-level", log_level_logfile] + self.args

        parser = self.parser()
        with patch(self.config_func, MagicMock(return_value=self.testing_config)):
            parser.parse_args(args)

        log_level_logfile_option = getattr(
            parser.options, self.logfile_loglevel_config_setting_name
        )

        # Check console logger
        self.assertEqual(self.log_impl.log_level_console, default_log_level)
        # Check extended logger
        self.assertEqual(
            self.log_impl.config[self.loglevel_config_setting_name],
            default_log_level,
        )
        self.assertEqual(
            self.log_impl.config[self.logfile_loglevel_config_setting_name],
            log_level_logfile,
        )
        # Check temp logger
        self.assertEqual(self.log_impl.temp_log_level, "error")
        # Check log file level setting
        self.assertEqual(log_level_logfile_option, log_level_logfile)
        # Check log file logger
        self.assertEqual(self.log_impl.log_level_logfile, log_level_logfile)

    def test_get_log_file_level_config(self):
        """
        Tests that log file level match the configured value
        """
        # Set defaults
        log_level = self.testing_config[self.loglevel_config_setting_name]

        args = self.args

        # Set log file level in config
        log_level_logfile = "info"
        opts = self.testing_config.copy()
        opts.update({self.logfile_loglevel_config_setting_name: log_level_logfile})

        parser = self.parser()
        with patch(self.config_func, MagicMock(return_value=opts)):
            parser.parse_args(args)

        log_level_logfile_option = getattr(
            parser.options, self.logfile_loglevel_config_setting_name
        )

        # Check console logger
        self.assertEqual(self.log_impl.log_level_console, log_level)
        # Check extended logger
        self.assertEqual(
            self.log_impl.config[self.loglevel_config_setting_name], log_level
        )
        self.assertEqual(
            self.log_impl.config[self.logfile_loglevel_config_setting_name],
            log_level_logfile,
        )
        # Check temp logger
        self.assertEqual(self.log_impl.temp_log_level, "error")
        # Check log file level setting
        self.assertEqual(log_level_logfile_option, log_level_logfile)
        # Check log file logger
        self.assertEqual(self.log_impl.log_level_logfile, log_level_logfile)

    def test_get_log_file_level_default(self):
        """
        Tests that log file level match the default value
        """
        # Set defaults
        default_log_level = self.testing_config[self.loglevel_config_setting_name]

        log_level = default_log_level
        log_level_logfile = default_log_level

        args = self.args

        parser = self.parser()
        with patch(self.config_func, MagicMock(return_value=self.testing_config)):
            parser.parse_args(args)

        log_level_logfile_option = getattr(
            parser.options, self.logfile_loglevel_config_setting_name
        )

        # Check console logger
        self.assertEqual(self.log_impl.log_level_console, log_level)
        # Check extended logger
        self.assertEqual(
            self.log_impl.config[self.loglevel_config_setting_name], log_level
        )
        self.assertEqual(
            self.log_impl.config[self.logfile_loglevel_config_setting_name],
            log_level_logfile,
        )
        # Check temp logger
        self.assertEqual(self.log_impl.temp_log_level, "error")
        # Check log file level setting
        self.assertEqual(log_level_logfile_option, log_level_logfile)
        # Check log file logger
        self.assertEqual(self.log_impl.log_level_logfile, log_level_logfile)
        # Check help message
        self.assertIn(
            "Default: '{}'.".format(default_log_level),
            parser.get_option("--log-file-level").help,
        )

    def test_get_console_log_level_with_file_log_level(
        self,
    ):  # pylint: disable=invalid-name
        """
        Tests that both console log level and log file level setting are working together
        """
        log_level = "critical"
        log_level_logfile = "debug"

        args = ["--log-file-level", log_level_logfile] + self.args

        opts = self.testing_config.copy()
        opts.update({self.loglevel_config_setting_name: log_level})

        parser = self.parser()
        with patch(self.config_func, MagicMock(return_value=opts)):
            parser.parse_args(args)

        log_level_logfile_option = getattr(
            parser.options, self.logfile_loglevel_config_setting_name
        )

        # Check console logger
        self.assertEqual(self.log_impl.log_level_console, log_level)
        # Check extended logger
        self.assertEqual(
            self.log_impl.config[self.loglevel_config_setting_name], log_level
        )
        self.assertEqual(
            self.log_impl.config[self.logfile_loglevel_config_setting_name],
            log_level_logfile,
        )
        # Check temp logger
        self.assertEqual(self.log_impl.temp_log_level, "error")
        # Check log file level setting
        self.assertEqual(log_level_logfile_option, log_level_logfile)
        # Check log file logger
        self.assertEqual(self.log_impl.log_level_logfile, log_level_logfile)

    def test_log_created(self):
        """
        Tests that log file is created
        """
        args = self.args
        log_file = self.log_file
        log_file_name = self.logfile_config_setting_name
        opts = self.testing_config.copy()
        opts.update({"log_file": log_file})
        if log_file_name != "log_file":
            opts.update({log_file_name: getattr(self, log_file_name)})

        parser = self.parser()
        with patch(self.config_func, MagicMock(return_value=opts)):
            parser.parse_args(args)

        if log_file_name == "log_file":
            self.assertGreaterEqual(os.path.getsize(log_file), 0)
        else:
            self.assertGreaterEqual(os.path.getsize(getattr(self, log_file_name)), 0)

    def test_callbacks_uniqueness(self):
        """
        Test that the callbacks are only added once, no matter
        how many instances of the parser we create
        """
        mixin_container_names = (
            "_mixin_setup_funcs",
            "_mixin_process_funcs",
            "_mixin_after_parsed_funcs",
            "_mixin_before_exit_funcs",
        )
        parser = self.parser()
        nums_1 = {}
        for cb_container in mixin_container_names:
            obj = getattr(parser, cb_container)
            nums_1[cb_container] = len(obj)

        # The next time we instantiate the parser, the counts should be equal
        parser = self.parser()
        nums_2 = {}
        for cb_container in mixin_container_names:
            obj = getattr(parser, cb_container)
            nums_2[cb_container] = len(obj)
        self.assertDictEqual(nums_1, nums_2)

    def test_verify_log_warning_logged(self):
        args = ["--log-level", "debug"] + self.args
        with TstSuiteLoggingHandler(level=logging.DEBUG) as handler:
            parser = self.parser()
            with patch(self.config_func, MagicMock(return_value=self.testing_config)):
                parser.parse_args(args)
            self.assertIn(
                "WARNING:Insecure logging configuration detected! Sensitive data may be logged.",
                handler.messages,
            )


class MasterOptionParserTestCase(ParserBase, TestCase):
    """
    Tests parsing Salt Master options
    """

    def setUp(self):
        """
        Setting up
        """
        # Set defaults
        self.default_config = salt.config.DEFAULT_MASTER_OPTS.copy()
        self.addCleanup(delattr, self, "default_config")

        # Log file
        # We need to use NamedTemporaryFile because Windows won't allow deleting
        # the log file even after it has been closed: WindowsError 32
        log_file = tempfile.NamedTemporaryFile(
            prefix="test_parsers_",
            suffix="_salt_master_parser_test",
            dir=RUNTIME_VARS.TMP,
            delete=True,
        )
        self.log_file = log_file.name
        log_file.close()
        # Function to patch
        self.config_func = "salt.config.master_config"

        # Mock log setup
        self.setup_log()

        # Assign parser
        self.parser = salt.utils.parsers.MasterOptionParser
        self.addCleanup(delattr, self, "parser")


class MinionOptionParserTestCase(ParserBase, TestCase):
    """
    Tests parsing Salt Minion options
    """

    def setUp(self):
        """
        Setting up
        """
        # Set defaults
        self.default_config = salt.config.DEFAULT_MINION_OPTS.copy()
        self.addCleanup(delattr, self, "default_config")

        # Log file
        # We need to use NamedTemporaryFile because Windows won't allow deleting
        # the log file even after it has been closed: WindowsError 32
        log_file = tempfile.NamedTemporaryFile(
            prefix="test_parsers_",
            suffix="_salt_minion_parser_test",
            dir=RUNTIME_VARS.TMP,
            delete=True,
        )
        self.log_file = log_file.name
        log_file.close()
        # Function to patch
        self.config_func = "salt.config.minion_config"

        # Mock log setup
        self.setup_log()

        # Assign parser
        self.parser = salt.utils.parsers.MinionOptionParser
        self.addCleanup(delattr, self, "parser")


class ProxyMinionOptionParserTestCase(ParserBase, TestCase):
    """
    Tests parsing Salt Proxy Minion options
    """

    def setUp(self):
        """
        Setting up
        """
        # Set defaults
        self.default_config = salt.config.DEFAULT_MINION_OPTS.copy()
        self.default_config.update(salt.config.DEFAULT_PROXY_MINION_OPTS)
        self.addCleanup(delattr, self, "default_config")

        # Log file
        # We need to use NamedTemporaryFile because Windows won't allow deleting
        # the log file even after it has been closed: WindowsError 32
        log_file = tempfile.NamedTemporaryFile(
            prefix="test_parsers_",
            suffix="_salt_proxy_minion_parser_test",
            dir=RUNTIME_VARS.TMP,
            delete=True,
        )
        self.log_file = log_file.name
        log_file.close()
        # Function to patch
        self.config_func = "salt.config.proxy_config"

        # Mock log setup
        self.setup_log()

        # Assign parser
        self.parser = salt.utils.parsers.ProxyMinionOptionParser
        self.addCleanup(delattr, self, "parser")


class SyndicOptionParserTestCase(ParserBase, TestCase):
    """
    Tests parsing Salt Syndic options
    """

    def setUp(self):
        """
        Setting up
        """
        # Set config option names
        self.logfile_config_setting_name = "syndic_log_file"

        # Set defaults
        self.default_config = salt.config.DEFAULT_MASTER_OPTS.copy()
        self.addCleanup(delattr, self, "default_config")

        # Log file
        # We need to use NamedTemporaryFile because Windows won't allow deleting
        # the log file even after it has been closed: WindowsError 32
        log_file = tempfile.NamedTemporaryFile(
            prefix="test_parsers_",
            suffix="_salt_syndic_parser_test",
            dir=RUNTIME_VARS.TMP,
            delete=True,
        )
        self.log_file = log_file.name
        log_file.close()
        syndic_log_file = tempfile.NamedTemporaryFile(
            prefix="test_parsers_",
            suffix="_salt_syndic_log",
            dir=RUNTIME_VARS.TMP,
            delete=True,
        )
        self.syndic_log_file = syndic_log_file.name
        syndic_log_file.close()
        # Function to patch
        self.config_func = "salt.config.syndic_config"

        # Mock log setup
        self.setup_log()

        # Assign parser
        self.parser = salt.utils.parsers.SyndicOptionParser
        self.addCleanup(delattr, self, "parser")


class SaltCMDOptionParserTestCase(ParserBase, TestCase):
    """
    Tests parsing Salt CLI options
    """

    def setUp(self):
        """
        Setting up
        """
        # Set mandatory CLI options
        self.args = ["foo", "bar.baz"]

        # Set defaults
        self.default_config = salt.config.DEFAULT_MASTER_OPTS.copy()
        self.addCleanup(delattr, self, "default_config")

        # Log file
        # We need to use NamedTemporaryFile because Windows won't allow deleting
        # the log file even after it has been closed: WindowsError 32
        log_file = tempfile.NamedTemporaryFile(
            prefix="test_parsers_",
            suffix="_salt_cmd_parser_test",
            dir=RUNTIME_VARS.TMP,
            delete=True,
        )
        self.log_file = log_file.name
        log_file.close()
        # Function to patch
        self.config_func = "salt.config.client_config"

        # Mock log setup
        self.setup_log()

        # Assign parser
        self.parser = salt.utils.parsers.SaltCMDOptionParser
        self.addCleanup(delattr, self, "parser")


class SaltCPOptionParserTestCase(ParserBase, TestCase):
    """
    Tests parsing salt-cp options
    """

    def setUp(self):
        """
        Setting up
        """
        # Set mandatory CLI options
        self.args = ["foo", "bar", "baz"]

        # Set defaults
        self.default_config = salt.config.DEFAULT_MASTER_OPTS.copy()
        self.addCleanup(delattr, self, "default_config")

        # Log file
        # We need to use NamedTemporaryFile because Windows won't allow deleting
        # the log file even after it has been closed: WindowsError 32
        log_file = tempfile.NamedTemporaryFile(
            prefix="test_parsers_",
            suffix="_salt_cp_parser_test",
            dir=RUNTIME_VARS.TMP,
            delete=True,
        )
        self.log_file = log_file.name
        log_file.close()
        # Function to patch
        self.config_func = "salt.config.master_config"

        # Mock log setup
        self.setup_log()

        # Assign parser
        self.parser = salt.utils.parsers.SaltCPOptionParser
        self.addCleanup(delattr, self, "parser")


class SaltKeyOptionParserTestCase(ParserBase, TestCase):
    """
    Tests parsing salt-key options
    """

    def setUp(self):
        """
        Setting up
        """
        # Set config option names
        self.logfile_config_setting_name = "key_logfile"

        # Set defaults
        self.default_config = salt.config.DEFAULT_MASTER_OPTS.copy()
        self.addCleanup(delattr, self, "default_config")

        # Log file
        # We need to use NamedTemporaryFile because Windows won't allow deleting
        # the log file even after it has been closed: WindowsError 32
        log_file = tempfile.NamedTemporaryFile(
            prefix="test_parsers_",
            suffix="_salt_key_parser_test",
            dir=RUNTIME_VARS.TMP,
            delete=True,
        )
        self.log_file = log_file.name
        log_file.close()
        key_logfile = tempfile.NamedTemporaryFile(
            prefix="test_parsers_",
            suffix="_key_logfile",
            dir=RUNTIME_VARS.TMP,
            delete=True,
        )
        self.key_logfile = key_logfile.name
        key_logfile.close()
        # Function to patch
        self.config_func = "salt.config.client_config"

        # Mock log setup
        self.setup_log()

        # Assign parser
        self.parser = salt.utils.parsers.SaltKeyOptionParser
        self.addCleanup(delattr, self, "parser")


class SaltCallOptionParserTestCase(ParserBase, TestCase):
    """
    Tests parsing Salt Minion options
    """

    def setUp(self):
        """
        Setting up
        """
        # Set mandatory CLI options
        self.args = ["foo.bar"]

        # Set defaults
        self.default_config = salt.config.DEFAULT_MINION_OPTS.copy()
        self.addCleanup(delattr, self, "default_config")

        # Log file
        # We need to use NamedTemporaryFile because Windows won't allow deleting
        # the log file even after it has been closed: WindowsError 32
        log_file = tempfile.NamedTemporaryFile(
            prefix="test_parsers_",
            suffix="_salt_call_parser_test",
            dir=RUNTIME_VARS.TMP,
            delete=True,
        )
        self.log_file = log_file.name
        log_file.close()
        # Function to patch
        self.config_func = "salt.config.minion_config"

        # Mock log setup
        self.setup_log()

        # Assign parser
        self.parser = salt.utils.parsers.SaltCallOptionParser
        self.addCleanup(delattr, self, "parser")


class SaltRunOptionParserTestCase(ParserBase, TestCase):
    """
    Tests parsing Salt Master options
    """

    def setUp(self):
        """
        Setting up
        """
        # Set mandatory CLI options
        self.args = ["foo.bar"]

        # Set defaults
        self.default_config = salt.config.DEFAULT_MASTER_OPTS.copy()
        self.addCleanup(delattr, self, "default_config")

        # Log file
        # We need to use NamedTemporaryFile because Windows won't allow deleting
        # the log file even after it has been closed: WindowsError 32
        log_file = tempfile.NamedTemporaryFile(
            prefix="test_parsers_",
            suffix="_salt_run_parser_test",
            dir=RUNTIME_VARS.TMP,
            delete=True,
        )
        self.log_file = log_file.name
        log_file.close()
        # Function to patch
        self.config_func = "salt.config.master_config"

        # Mock log setup
        self.setup_log()

        # Assign parser
        self.parser = salt.utils.parsers.SaltRunOptionParser
        self.addCleanup(delattr, self, "parser")


class SaltSSHOptionParserTestCase(ParserBase, TestCase):
    """
    Tests parsing Salt Master options
    """

    def setUp(self):
        """
        Setting up
        """
        # Set mandatory CLI options
        self.args = ["foo", "bar.baz"]

        # Set config option names
        self.logfile_config_setting_name = "ssh_log_file"

        # Set defaults
        self.default_config = salt.config.DEFAULT_MASTER_OPTS.copy()
        self.addCleanup(delattr, self, "default_config")

        # Log file
        # We need to use NamedTemporaryFile because Windows won't allow deleting
        # the log file even after it has been closed: WindowsError 32
        log_file = tempfile.NamedTemporaryFile(
            prefix="test_parsers_",
            suffix="_salt_ssh_parser_test",
            dir=RUNTIME_VARS.TMP,
            delete=True,
        )
        self.log_file = log_file.name
        log_file.close()
        ssh_log_file = tempfile.NamedTemporaryFile(
            prefix="test_parsers_",
            suffix="_ssh_logfile",
            dir=RUNTIME_VARS.TMP,
            delete=True,
        )
        self.ssh_log_file = ssh_log_file.name
        ssh_log_file.close()
        # Function to patch
        self.config_func = "salt.config.master_config"

        # Mock log setup
        self.setup_log()

        # Assign parser
        self.parser = salt.utils.parsers.SaltSSHOptionParser
        self.addCleanup(delattr, self, "parser")


class SaltCloudParserTestCase(ParserBase, TestCase):
    """
    Tests parsing Salt Cloud options
    """

    def setUp(self):
        """
        Setting up
        """
        # Set mandatory CLI options
        self.args = ["-p", "foo", "bar"]

        # Set default configs
        # Cloud configs are merged with master configs in
        # config/__init__.py, so we'll do that here as well
        # As we need the 'user' key later on.
        self.default_config = salt.config.DEFAULT_MASTER_OPTS.copy()
        self.default_config.update(salt.config.DEFAULT_CLOUD_OPTS)
        self.addCleanup(delattr, self, "default_config")

        # Log file
        # We need to use NamedTemporaryFile because Windows won't allow deleting
        # the log file even after it has been closed: WindowsError 32
        log_file = tempfile.NamedTemporaryFile(
            prefix="test_parsers_",
            suffix="_salt_cloud_parser_test",
            dir=RUNTIME_VARS.TMP,
            delete=True,
        )
        self.log_file = log_file.name
        log_file.close()
        # Function to patch
        self.config_func = "salt.config.cloud_config"

        # Mock log setup
        self.setup_log()

        # Assign parser
        self.parser = salt.utils.parsers.SaltCloudParser
        self.addCleanup(delattr, self, "parser")


class SPMParserTestCase(ParserBase, TestCase):
    """
    Tests parsing Salt Cloud options
    """

    def setUp(self):
        """
        Setting up
        """
        # Set mandatory CLI options
        self.args = ["foo", "bar"]

        # Set config option names
        self.logfile_config_setting_name = "spm_logfile"

        # Set defaults
        self.default_config = salt.config.DEFAULT_MASTER_OPTS.copy()
        self.default_config.update(salt.config.DEFAULT_SPM_OPTS)
        self.addCleanup(delattr, self, "default_config")

        # Log file
        # We need to use NamedTemporaryFile because Windows won't allow deleting
        # the log file even after it has been closed: WindowsError 32
        log_file = tempfile.NamedTemporaryFile(
            prefix="test_parsers_",
            suffix="_spm_parser_test",
            dir=RUNTIME_VARS.TMP,
            delete=True,
        )
        self.log_file = log_file.name
        log_file.close()
        spm_logfile = tempfile.NamedTemporaryFile(
            prefix="test_parsers_",
            suffix="_spm_logfile",
            dir=RUNTIME_VARS.TMP,
            delete=True,
        )
        self.spm_logfile = spm_logfile.name
        spm_logfile.close()
        # Function to patch
        self.config_func = "salt.config.spm_config"

        # Mock log setup
        self.setup_log()

        # Assign parser
        self.parser = salt.utils.parsers.SPMParser
        self.addCleanup(delattr, self, "parser")


class SaltAPIParserTestCase(ParserBase, TestCase):
    """
    Tests parsing Salt Cloud options
    """

    def setUp(self):
        """
        Setting up
        """
        # Set mandatory CLI options
        self.args = []

        # Set config option names
        self.logfile_config_setting_name = "api_logfile"

        # Set defaults
        self.default_config = salt.config.DEFAULT_MASTER_OPTS.copy()
        self.default_config.update(salt.config.DEFAULT_API_OPTS)
        self.addCleanup(delattr, self, "default_config")

        # Log file
        # We need to use NamedTemporaryFile because Windows won't allow deleting
        # the log file even after it has been closed: WindowsError 32
        log_file = tempfile.NamedTemporaryFile(
            prefix="test_parsers_",
            suffix="_salt_api_parser_test",
            dir=RUNTIME_VARS.TMP,
            delete=True,
        )
        self.log_file = log_file.name
        log_file.close()
        api_logfile = tempfile.NamedTemporaryFile(
            prefix="test_parsers_",
            suffix="_api_logfile",
            dir=RUNTIME_VARS.TMP,
            delete=True,
        )
        self.api_logfile = api_logfile.name
        api_logfile.close()
        # Function to patch
        self.config_func = "salt.config.api_config"

        # Mock log setup
        self.setup_log()

        # Assign parser
        self.parser = salt.utils.parsers.SaltAPIParser
        self.addCleanup(delattr, self, "parser")


class DaemonMixInTestCase(TestCase):
    """
    Tests the PIDfile deletion in the DaemonMixIn.
    """

    def setUp(self):
        """
        Setting up
        """
        # Setup mixin
        self.daemon_mixin = salt.utils.parsers.DaemonMixIn()
        self.daemon_mixin.config = {}
        self.daemon_mixin.config["pidfile"] = "/some/fake.pid"

    def tearDown(self):
        """
        Tear down test
        :return:
        """
        del self.daemon_mixin

    @patch("os.unlink", MagicMock())
    @patch("os.path.isfile", MagicMock(return_value=True))
    @patch("salt.utils.parsers.log", MagicMock())
    def test_pid_file_deletion(self):
        """
        PIDfile deletion without exception.
        """
        self.daemon_mixin._mixin_before_exit()
        assert salt.utils.parsers.os.unlink.call_count == 1
        salt.utils.parsers.log.info.assert_not_called()
        salt.utils.parsers.log.debug.assert_not_called()

    @patch("os.unlink", MagicMock(side_effect=OSError()))
    @patch("os.path.isfile", MagicMock(return_value=True))
    @patch("salt.utils.parsers.log", MagicMock())
    def test_pid_deleted_oserror_as_root(self):
        """
        PIDfile deletion with exception, running as root.
        """
        if salt.utils.platform.is_windows():
            patch_args = (
                "salt.utils.win_functions.is_admin",
                MagicMock(return_value=True),
            )
        else:
            patch_args = ("os.getuid", MagicMock(return_value=0))

        with patch(*patch_args):
            self.daemon_mixin._mixin_before_exit()
            assert salt.utils.parsers.os.unlink.call_count == 1
            salt.utils.parsers.log.info.assert_called_with(
                "PIDfile(%s) could not be deleted: %s",
                format(self.daemon_mixin.config["pidfile"], ""),
                ANY,
                exc_info_on_loglevel=logging.DEBUG,
            )

    @patch("os.unlink", MagicMock(side_effect=OSError()))
    @patch("os.path.isfile", MagicMock(return_value=True))
    @patch("salt.utils.parsers.log", MagicMock())
    def test_pid_deleted_oserror_as_non_root(self):
        """
        PIDfile deletion with exception, running as non-root.
        """
        if salt.utils.platform.is_windows():
            patch_args = (
                "salt.utils.win_functions.is_admin",
                MagicMock(return_value=False),
            )
        else:
            patch_args = ("os.getuid", MagicMock(return_value=1000))

        with patch(*patch_args):
            self.daemon_mixin._mixin_before_exit()
            assert salt.utils.parsers.os.unlink.call_count == 1
            salt.utils.parsers.log.info.assert_not_called()
            salt.utils.parsers.log.debug.assert_not_called()
